/*
 *************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002-2007, Ralink Technology, Inc.
 *
 * This program is free software; you can redistribute it and/or modify  * 
 * it under the terms of the GNU General Public License as published by  * 
 * the Free Software Foundation; either version 2 of the License, or     * 
 * (at your option) any later version.                                   * 
 *                                                                       * 
 * This program is distributed in the hope that it will be useful,       * 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        * 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         * 
 * GNU General Public License for more details.                          * 
 *                                                                       * 
 * You should have received a copy of the GNU General Public License     * 
 * along with this program; if not, write to the                         * 
 * Free Software Foundation, Inc.,                                       * 
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             * 
 *                                                                       * 
 *************************************************************************
 
	Module Name:
	rtmp_data.c

	Abstract:
	Data path subroutines

	Revision History:
	Who 		When			What
	--------	----------		----------------------------------------------
	John		      Aug/17/04		major modification for RT2561/2661
	Jan Lee	      Mar/17/06		major modification for RT2860 New Ring Design
*/
#include "rt_config.h"


VOID STARxEAPOLFrameIndicate(
	IN	PRTMP_ADAPTER	pAd,
	IN	MAC_TABLE_ENTRY	*pEntry,
	IN	RX_BLK			*pRxBlk,
	IN	UCHAR			FromWhichBSSID)
{
	PRXD_STRUC		pRxD = &(pRxBlk->RxD);
	PRXWI_STRUC		pRxWI = pRxBlk->pRxWI;
	UCHAR			*pTmpBuf;

#ifdef WPA_SUPPLICANT_SUPPORT
	if (pAd->StaCfg.WpaSupplicantUP)
	{
		// All EAPoL frames have to pass to upper layer (ex. WPA_SUPPLICANT daemon)
		// TBD : process fragmented EAPol frames
		{
			// In 802.1x mode, if the received frame is EAP-SUCCESS packet, turn on the PortSecured variable
			if ( pAd->StaCfg.IEEE8021X == TRUE &&
				 (EAP_CODE_SUCCESS == WpaCheckEapCode(pAd, pRxBlk->pData, pRxBlk->DataSize, LENGTH_802_1_H)))
			{
				PUCHAR	Key; 			
				UCHAR 	CipherAlg;
				int     idx = 0;
				MAC_TABLE_ENTRY *pEntry = &pAd->MacTab.Content[BSSID_WCID];

				DBGPRINT_RAW(RT_DEBUG_TRACE, ("Receive EAP-SUCCESS Packet\n"));
				pAd->StaCfg.PortSecured = WPA_802_1X_PORT_SECURED;	
				
                if (pAd->StaCfg.IEEE8021x_required_keys == FALSE)
                {
                    idx = pAd->StaCfg.DesireSharedKeyId;
                    CipherAlg = pAd->StaCfg.DesireSharedKey[idx].CipherAlg;
					Key = pAd->StaCfg.DesireSharedKey[idx].Key;
                    
                    if (pAd->StaCfg.DesireSharedKey[idx].KeyLen > 0)
    				{
						// Set key material and cipherAlg to Asic
						AsicAddSharedKeyEntry(pAd, BSS0, idx, CipherAlg, Key, NULL, NULL);

						// Assign group key info  
						RTMPAddWcidAttributeEntry(pAd, BSS0, idx, CipherAlg, NULL);

						// Assign pairwise key info
						RTMPAddWcidAttributeEntry(pAd, BSS0, idx, CipherAlg, pEntry);
    				}
                }
			}

			Indicate_Legacy_Packet(pAd, pRxBlk, FromWhichBSSID);
			return; 
		}
	}
	else
#endif // WPA_SUPPLICANT_SUPPORT //                        
	{
		// Special DATA frame that has to pass to MLME
		//	 1. Cisco Aironet frames for CCX2. We need pass it to MLME for special process
		//	 2. EAPOL handshaking frames when driver supplicant enabled, pass to MLME for special process
		{
			pTmpBuf = pRxBlk->pData - LENGTH_802_11;
			NdisMoveMemory(pTmpBuf, pRxBlk->pHeader, LENGTH_802_11);
			REPORT_MGMT_FRAME_TO_MLME(pAd, pRxWI->WirelessCliID, pTmpBuf, pRxBlk->DataSize + LENGTH_802_11, pRxWI->RSSI0, pRxWI->RSSI1, pRxWI->RSSI2, pRxD->PlcpSignal);
			DBGPRINT_RAW(RT_DEBUG_TRACE, ("!!! report EAPOL/AIRONET DATA to MLME (len=%d) !!!\n", pRxBlk->DataSize));
		}
	}

	RELEASE_NDIS_PACKET(pAd, pRxBlk->pRxPacket, NDIS_STATUS_FAILURE);
	return;

}

VOID STARxDataFrameAnnounce(
	IN	PRTMP_ADAPTER	pAd,
	IN	MAC_TABLE_ENTRY	*pEntry,
	IN	RX_BLK			*pRxBlk,
	IN	UCHAR			FromWhichBSSID)
{

	// non-EAP frame
	if (!RTMPCheckWPAframe(pAd, pRxBlk->pData, pRxBlk->DataSize))
	{
		// drop all non-EAP DATA frame before
		// this client's Port-Access-Control is secured
		if (pRxBlk->pHeader->FC.Wep)
		{
			// unsupported cipher suite
			if (pAd->StaCfg.WepStatus == Ndis802_11EncryptionDisabled)
			{
				// release packet
				RELEASE_NDIS_PACKET(pAd, pRxBlk->pRxPacket, NDIS_STATUS_FAILURE);
				return;
			}
		}
		else
		{
			// encryption in-use but receive a non-EAPOL clear text frame, drop it 
			if ((pAd->StaCfg.WepStatus != Ndis802_11EncryptionDisabled) && 
				(pAd->StaCfg.PortSecured == WPA_802_1X_PORT_NOT_SECURED))
			{
				// release packet
				RELEASE_NDIS_PACKET(pAd, pRxBlk->pRxPacket, NDIS_STATUS_FAILURE);
				return;
			}
		}

		if (!RX_BLK_TEST_FLAG(pRxBlk, fRX_ARALINK))
		{
			// Normal legacy, AMPDU or AMSDU
			CmmRxnonRalinkFrameIndicate(pAd, pEntry, pRxBlk, FromWhichBSSID);
			
		}
		else
		{
			// ARALINK
			CmmRxRalinkFrameIndicate(pAd, pEntry, pRxBlk, FromWhichBSSID);
		}
	}
	else 
	{
		// Determin the destination of the EAP frame
		//  to WPA state machine or upper layer
		STARxEAPOLFrameIndicate(pAd, pEntry, pRxBlk, FromWhichBSSID);
	}
}


// For TKIP frame, calculate the MIC value						
BOOLEAN STACheckTkipMICValue(
	IN	PRTMP_ADAPTER	pAd,
	IN	MAC_TABLE_ENTRY	*pEntry,
	IN	RX_BLK			*pRxBlk)
{
	PHEADER_802_11	pHeader = pRxBlk->pHeader;
	UCHAR			*pData = pRxBlk->pData;
	USHORT			DataSize = pRxBlk->DataSize;
	PCIPHER_KEY		pWpaKey;
	UCHAR			*pDA, *pSA;

	pWpaKey = &pAd->SharedKey[BSS0][pRxBlk->pRxWI->KeyIndex];

	// Minus MIC length
	DataSize -= 8;

	pDA = pHeader->Addr1;
	if (RX_BLK_TEST_FLAG(pRxBlk, fRX_INFRA))
	{
		pSA = pHeader->Addr3;
	}
	else 
	{
		pSA = pHeader->Addr2;
	}

	if (RTMPTkipCompareMICValue(pAd,
								pData,
								pDA,
								pSA,
								pWpaKey->RxMic,
								DataSize) == FALSE)
	{
		DBGPRINT_RAW(RT_DEBUG_ERROR,("Rx MIC Value error 2\n"));

#ifdef WPA_SUPPLICANT_SUPPORT
		if (pAd->StaCfg.WpaSupplicantUP)
		{
			WpaSendMicFailureToWpaSupplicant(pAd, (pWpaKey->Type == PAIRWISEKEY) ? TRUE : FALSE);
		}
		else 
#endif // WPA_SUPPLICANT_SUPPORT //
		{
			RTMPReportMicError(pAd, pWpaKey);
		}
		
		// release packet
		RELEASE_NDIS_PACKET(pAd, pRxBlk->pRxPacket, NDIS_STATUS_FAILURE);
		return FALSE;
	}

	return TRUE;
}


//
// All Rx routines use RX_BLK structure to hande rx events
// It is very important to build pRxBlk attributes
//  1. pHeader pointer to 802.11 Header
//  2. pData pointer to payload including LLC (just skip Header)
//  3. set payload size including LLC to DataSize
//  4. set some flags with RX_BLK_SET_FLAG()
// 
VOID STAHandleRxDataFrame(
	IN	PRTMP_ADAPTER	pAd,
	IN	RX_BLK			*pRxBlk)
{
	PRXD_STRUC						pRxD = &(pRxBlk->RxD);
	PRXWI_STRUC						pRxWI = pRxBlk->pRxWI;
	PHEADER_802_11					pHeader = pRxBlk->pHeader;
	PNDIS_PACKET					pRxPacket = pRxBlk->pRxPacket;
	BOOLEAN 						bFragment = FALSE;
	MAC_TABLE_ENTRY	    			*pEntry = NULL;
	UCHAR							FromWhichBSSID = BSS0;

	// before LINK UP, all DATA frames are rejected
	if (!OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_MEDIA_STATE_CONNECTED))
	{
		// release packet
		RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
		return;
	}

	// Drop not my BSS frames
	if (pRxD->MyBss == 0)
	{
		// release packet
		RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
		return;
	}

	if (pAd->CommonCfg.bAPSDCapable && pAd->CommonCfg.APEdcaParm.bAPSDCapable && (pHeader->FC.SubType & 0x08))
	{
		UCHAR *pData;
		DBGPRINT(RT_DEBUG_TRACE,("bAPSDCapable\n"));

		// Qos bit 4
		pData = (PUCHAR)pHeader + LENGTH_802_11;
		if ((*pData >> 4) & 0x01)
		{
			DBGPRINT(RT_DEBUG_TRACE,("RxDone- Rcv EOSP frame, driver may fall into sleep\n"));
			pAd->CommonCfg.bInServicePeriod = FALSE;

			// Force driver to fall into sleep mode when rcv EOSP frame
			if (!OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_DOZE))
			{
				USHORT  TbttNumToNextWakeUp;
				USHORT  NextDtim = pAd->StaCfg.DtimPeriod;
				ULONG   Now;

				NdisGetSystemUpTime(&Now);
				NextDtim -= (USHORT)(Now - pAd->StaCfg.LastBeaconRxTime)/pAd->CommonCfg.BeaconPeriod;

				TbttNumToNextWakeUp = pAd->StaCfg.DefaultListenCount;
				if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_RECEIVE_DTIM) && (TbttNumToNextWakeUp > NextDtim))
					TbttNumToNextWakeUp = NextDtim;

				MlmeSetPsmBit(pAd, PWR_SAVE);
				// if WMM-APSD is failed, try to disable following line
				AsicSleepThenAutoWakeup(pAd, TbttNumToNextWakeUp);
			}
		}

		if ((pHeader->FC.MoreData) && (pAd->CommonCfg.bInServicePeriod))
		{
			DBGPRINT(RT_DEBUG_TRACE,("Sending another trigger frame when More Data bit is set to 1\n"));
			//STA should not send trigger frame, the trigger frame should be level 3.						
			//RTMPSendNullFrame(pAd, pAd->CommonCfg.TxRate, TRUE);
		}
	}

	// Drop NULL, CF-ACK(no data), CF-POLL(no data), and CF-ACK+CF-POLL(no data) data frame
	if ((pHeader->FC.SubType & 0x04)) // bit 2 : no DATA
	{
		// release packet
		RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
		return;
	}

    // Drop not my BSS frame (we can not only check the MyBss bit in RxD)
	if (INFRA_ON(pAd))
	{
		// Infrastructure mode, check address 2 for BSSID
		if (!RTMPEqualMemory(&pHeader->Addr2, &pAd->CommonCfg.Bssid, 6))
		{
			// Receive frame not my BSSID
            // release packet
            RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			return;
		}
	}
	else	// Ad-Hoc mode or Not associated
	{
		// Ad-Hoc mode, check address 3 for BSSID
		if (!RTMPEqualMemory(&pHeader->Addr3, &pAd->CommonCfg.Bssid, 6))
		{
			// Receive frame not my BSSID
            // release packet
            RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			return;
		}
	}

	//
	// find pEntry
	// 
	if (pRxWI->WirelessCliID < MAX_LEN_OF_MAC_TABLE)
	{
		pEntry = &pAd->MacTab.Content[pRxWI->WirelessCliID];
	}
	else 
	{
		// 1. release packet if infra mode 
		// 2. new a pEntry if ad-hoc mode
		RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
		return;
	}

	// infra or ad-hoc
	if (INFRA_ON(pAd))
	{
		RX_BLK_SET_FLAG(pRxBlk, fRX_INFRA);
		ASSERT(pRxWI->WirelessCliID == BSSID_WCID);

	}
	
	// check Atheros Client
	if ((pEntry->bIAmBadAtheros == FALSE) &&  (pRxD->AMPDU == 1) && (pHeader->FC.Retry ))
	{
		pEntry->bIAmBadAtheros = TRUE;
		pAd->CommonCfg.IOTestParm.bCurrentAtheros = TRUE;
		pAd->CommonCfg.IOTestParm.bLastAtheros = TRUE;
		if (!STA_AES_ON(pAd))
		{
			AsicUpdateProtect(pAd, 8, ALLN_SETPROTECT, TRUE, FALSE);
		}
		DBGPRINT(RT_DEBUG_INFO, ("Atheros Problem. Turn on RTS/CTS!!!\n"));
	}


	pRxBlk->pData = (UCHAR *)pHeader;

	//
	// update RxBlk->pData, DataSize
	// 802.11 Header, QOS, HTC, Hw Padding
	// 

	// 1. skip 802.11 HEADER
	pRxBlk->pData += LENGTH_802_11;
	pRxBlk->DataSize -= LENGTH_802_11;

	// 2. QOS
	if (pHeader->FC.SubType & 0x08)
	{
		RX_BLK_SET_FLAG(pRxBlk, fRX_QOS);

		// bit 7 in QoS Control field signals the HT A-MSDU format
		if ((*pRxBlk->pData) & 0x80)
		{
			RX_BLK_SET_FLAG(pRxBlk, fRX_AMSDU);
		}

		// skip QOS contorl field
		pRxBlk->pData += 2;
		pRxBlk->DataSize -=2;
	}

	// 3. Order bit: A-Ralink or HTC+
	if (pHeader->FC.Order)
	{
#ifdef AGGREGATION_SUPPORT
		if ((pRxWI->PHYMODE < MODE_HTMIX) && (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED)))
		{
			RX_BLK_SET_FLAG(pRxBlk, fRX_ARALINK);
		}
		else
#endif
		{
			RX_BLK_SET_FLAG(pRxBlk, fRX_HTC);
			// skip HTC contorl field
			pRxBlk->pData += 4;
			pRxBlk->DataSize -= 4;
		}
	}

	// 4. skip HW padding 
	if (pRxD->L2PAD)
	{
		// just move pData pointer 
		// because DataSize excluding HW padding 
		RX_BLK_SET_FLAG(pRxBlk, fRX_PAD);
		pRxBlk->pData += 2;
	}

	if (pRxD->BA)
	{
		RX_BLK_SET_FLAG(pRxBlk, fRX_AMPDU);
	}

	// 
	// Case I  Process Broadcast & Multicast data frame
	//
	if (pRxD->Bcast || pRxD->Mcast)
	{
		INC_COUNTER64(pAd->WlanCounters.MulticastReceivedFrameCount);

		// Drop Mcast/Bcast frame with fragment bit on
		if (pHeader->FC.MoreFrag)
		{
			// release packet
			RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			return;
		}
				
		// Filter out Bcast frame which AP relayed for us
		if (pHeader->FC.FrDs && MAC_ADDR_EQUAL(pHeader->Addr3, pAd->CurrentAddress))
		{
			// release packet
			RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			return;
		}

		Indicate_Legacy_Packet(pAd, pRxBlk, FromWhichBSSID);
		DBGPRINT_RAW(RT_DEBUG_INFO, ("!!! report BCAST DATA to LLC (len=%d) !!!\n", pRxBlk->DataSize));
		return;
	}					
	else if (pRxD->U2M)
	{
		// Send PS-Poll for AP to send next data frame					
		// if ((pHeader->FC.MoreData) && INFRA_ON(pAd) && (pAd->StaCfg.Psm == PWR_SAVE))
		// EnqueuePsPoll(pAd);					
		pAd->LastRxRate = (USHORT)((pRxWI->MCS) + (pRxWI->BW <<7) + (pRxWI->ShortGI <<8)+ (pRxWI->PHYMODE <<14)) ;

		Update_Rssi_Sample(pAd, &pAd->StaCfg.RssiSample, pRxWI);

		pAd->StaCfg.LastSNR0 = (UCHAR)(pRxWI->SNR0);
		pAd->StaCfg.LastSNR1 = (UCHAR)(pRxWI->SNR1);
	
		pAd->RalinkCounters.OneSecRxOkDataCnt++;


    	if (!((pHeader->Frag == 0) && (pHeader->FC.MoreFrag == 0)))
    	{
    		// re-assemble the fragmented packets
    		// return complete frame (pRxPacket) or NULL   
    		bFragment = TRUE;
    		pRxPacket = RTMPDeFragmentDataFrame(pAd, pRxBlk);
    	}

    	if (pRxPacket)
    	{
			pEntry = &pAd->MacTab.Content[pRxWI->WirelessCliID];

    		// process complete frame
    		if (bFragment && (pRxD->Decrypted) && (pEntry->WepStatus == Ndis802_11Encryption2Enabled))
    		{
    			// For TKIP frame, calculate the MIC value						
    			if (STACheckTkipMICValue(pAd, pEntry, pRxBlk) == FALSE)
    			{
    				return;
    			}
    		}

    		STARxDataFrameAnnounce(pAd, pEntry, pRxBlk, FromWhichBSSID);
			return;
    	}
    	else
    	{
    		// just return 
    		// because RTMPDeFragmentDataFrame() will release rx packet, 
    		// if packet is fragmented
    		return;
    	}
	}

	ASSERT(0);
	// release packet
	RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
}

VOID STAHandleRxMgmtFrame(
	IN	PRTMP_ADAPTER	pAd,
	IN	RX_BLK			*pRxBlk)
{
	PRXD_STRUC		pRxD = &(pRxBlk->RxD);
	PRXWI_STRUC		pRxWI = pRxBlk->pRxWI;
	PHEADER_802_11	pHeader = pRxBlk->pHeader;
	PNDIS_PACKET	pRxPacket = pRxBlk->pRxPacket;

	REPORT_MGMT_FRAME_TO_MLME(pAd, pRxWI->WirelessCliID, pHeader, pRxWI->MPDUtotalByteCount, 
								  pRxWI->RSSI0, pRxWI->RSSI1, pRxWI->RSSI2, pRxD->PlcpSignal);

	RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_SUCCESS);
}

VOID STAHandleRxControlFrame(
	IN	PRTMP_ADAPTER	pAd,
	IN	RX_BLK			*pRxBlk)
{
	PRXWI_STRUC		pRxWI = pRxBlk->pRxWI;
	PHEADER_802_11	pHeader = pRxBlk->pHeader;
	PNDIS_PACKET	pRxPacket = pRxBlk->pRxPacket;

	switch (pHeader->FC.SubType)
	{
		case SUBTYPE_BLOCK_ACK_REQ:
			{
				CntlEnqueueForRecv(pAd, pRxWI->WirelessCliID, (pRxWI->MPDUtotalByteCount), (PFRAME_BA_REQ)pHeader);
			}
			break;
		case SUBTYPE_BLOCK_ACK:
		case SUBTYPE_ACK:
		default:		
			DBGPRINT(RT_DEBUG_INFO,("ignore CNTL (subtype=%d)\n", pHeader->FC.SubType));
			break;
	}

	RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
}


/*
	========================================================================

	Routine Description:
		Process RxDone interrupt, running in DPC level

	Arguments:
		pAd Pointer to our adapter

	Return Value:
		None

	IRQL = DISPATCH_LEVEL
	
	Note:
		This routine has to maintain Rx ring read pointer.
		Need to consider QOS DATA format when converting to 802.3
	========================================================================
*/

BOOLEAN STARxDoneInterruptHandle(
	IN	PRTMP_ADAPTER	pAd) 
{
	UINT32			RxProcessed, RxPending;
	BOOLEAN			bReschedule = FALSE;
	RXD_STRUC		*pRxD;
	UCHAR			*pData;
	PRXWI_STRUC		pRxWI;
	PNDIS_PACKET	pRxPacket;
	PHEADER_802_11	pHeader;
	RX_BLK			RxCell;


	RxProcessed = RxPending = 0;

	// process whole rx ring
	while (1)
	{

		if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF |
								fRTMP_ADAPTER_RESET_IN_PROGRESS |
								fRTMP_ADAPTER_HALT_IN_PROGRESS) ||
			!RTMP_TEST_FLAG(pAd,fRTMP_ADAPTER_START_UP))
		{
			break;
		}

		if (RxProcessed++ > MAX_RX_PROCESS_CNT)
		{
			// need to reschedule rx handle 
			bReschedule = TRUE;
			break;
		}

		
		// 1. allocate a new data packet into rx ring to replace received packet 
		//    then processing the received packet
		// 2. the callee must take charge of release of packet
		// 3. As far as driver is concerned ,
		//    the rx packet must 
		//      a. be indicated to upper layer or 
		//      b. be released if it is discarded
		pRxPacket = GetPacketFromRxRing(pAd, &(RxCell.RxD), &bReschedule, &RxPending);
		if (pRxPacket == NULL)
		{
			// no more packet to process
			break;
		}

		// get rx ring descriptor
		pRxD = &(RxCell.RxD);
		// get rx data buffer
		pData	= GET_OS_PKT_DATAPTR(pRxPacket);
		pRxWI	= (PRXWI_STRUC) pData;
		pHeader = (PHEADER_802_11) (pData+RXWI_SIZE) ;



#ifdef BIG_ENDIAN
		RTMPFrameEndianChange(pAd, (PUCHAR)pHeader, DIR_READ, TRUE);
		RTMPWIEndianChange((PUCHAR)pRxWI, TYPE_RXWI);
#endif

		// build RxCell
		RxCell.pRxWI = pRxWI;
		RxCell.pHeader = pHeader;
		RxCell.pRxPacket = pRxPacket;
		RxCell.pData = (UCHAR *) pHeader;
		RxCell.DataSize = pRxWI->MPDUtotalByteCount;
		RxCell.Flags = 0;

		// Increase Total receive byte counter after real data received no mater any error or not
		pAd->RalinkCounters.ReceivedByteCount +=  pRxWI->MPDUtotalByteCount;
		pAd->RalinkCounters.RxCount ++;

		INC_COUNTER64(pAd->WlanCounters.ReceivedFragmentCount);

        if (MONITOR_ON(pAd))
		{
            send_monitor_packets(pAd, &RxCell);
            //RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			break;
		}

#ifdef RALINK_ATE
		if (pAd->ate.Mode != ATE_STOP)
		{
			pAd->ate.RxCntPerSec++;
			ATESampleRssi(pAd, pRxWI);
#ifdef RALINK_2860_QA
			if (pAd->ate.bQARxStart == TRUE)
			{
				ATE_QA_Statistics(pAd, pRxWI, pRxD,	pHeader);
			}
#endif // RALINK_2860_QA //
			RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_SUCCESS);
			continue;
		}
#endif // RALINK_ATE //

		// Check for all RxD errors
		if (RTMPCheckRxError(pAd, pHeader, pRxWI, pRxD) != NDIS_STATUS_SUCCESS)
		{
			pAd->Counters8023.RxErrors++;
			// discard this frame
			RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
			continue;
		}

		
		switch (pHeader->FC.Type)
		{
			// CASE I, receive a DATA frame
			case BTYPE_DATA:
				{
					// process DATA frame
					STAHandleRxDataFrame(pAd, &RxCell);
				}
				break;
			// CASE II, receive a MGMT frame
			case BTYPE_MGMT:
				{
					STAHandleRxMgmtFrame(pAd, &RxCell);
				}
				break;
			// CASE III. receive a CNTL frame
			case BTYPE_CNTL:
				{
					STAHandleRxControlFrame(pAd, &RxCell);
				}
				break;
			// discard other type
			default:
				RELEASE_NDIS_PACKET(pAd, pRxPacket, NDIS_STATUS_FAILURE);
				break;
		}
	}

	return bReschedule;
}


/*
	========================================================================

	Routine Description:
		Process TxDone interrupt, running in	DPC level

	Arguments:
		Adapter 	Pointer to our adapter

	Return Value:
		None

	IRQL = DISPATCH_LEVEL
	
	Note:

	========================================================================
*/
VOID	RTMPHandleTxDoneInterrupt(
	IN	PRTMP_ADAPTER	pAd)
{
}

// this portion already move to cmm_data.c
// RTMPHandleTxRingDmaDoneInterrupt,
// RTMPHandleMgmtRingDmaDoneInterrupt,
// RTMPHandleTBTTInterrupt,
// RTMPHandlePreTBTTInterrupt
// move to cmm_data.c

/*
	========================================================================

	Routine Description:
	Arguments:
		pAd 	Pointer to our adapter

	IRQL = DISPATCH_LEVEL
	
	========================================================================
*/
VOID	RTMPHandleTwakeupInterrupt(
	IN PRTMP_ADAPTER pAd)
{
	DBGPRINT(RT_DEBUG_INFO, ("Twakeup Expired... !!!\n"));
	AsicForceWakeup(pAd);
}

// this portion already move to cmm_data.c
// RTMPHandleRxCoherentInterrupt,
// MiniportMMRequest,
// MlmeHardTransmit,
// TxFrameIsAggregatible,
// Sniff2BytesFromNdisBuffer,
// RTMPDeQueuePacket


/*
	========================================================================

	Routine Description:
		This routine classifies outgoing frames into several AC (Access
		Category) and enqueue them into corresponding s/w waiting queues.
		
	Arguments:
		pAd Pointer to our adapter
		pPacket 	Pointer to send packet
		
	Return Value:
		None

	IRQL = DISPATCH_LEVEL
	
	Note:
	
	========================================================================
*/
NDIS_STATUS RTMPSendPacket(
	IN	PRTMP_ADAPTER	pAd,
	IN	PNDIS_PACKET	pPacket)
{
	PACKET_INFO 	PacketInfo;
	PUCHAR			pSrcBufVA;
	UINT			SrcBufLen;
	UINT			AllowFragSize;
	UCHAR			NumberOfFrag;
//	UCHAR			RTSRequired;
	UCHAR			QueIdx, UserPriority;

	// Prepare packet information structure for buffer descriptor 
	// chained within a single NDIS packet.
	RTMP_QueryPacketInfo(pPacket, &PacketInfo, &pSrcBufVA, &SrcBufLen);

	if (pSrcBufVA == NULL)
	{
		DBGPRINT(RT_DEBUG_ERROR,("RTMPSendPacket --> pSrcBufVA == NULL !!!SrcBufLen=%x\n",SrcBufLen));
		// Resourece is low, system did not allocate virtual address
		// return NDIS_STATUS_FAILURE directly to upper layer
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return NDIS_STATUS_FAILURE;
	}

	DBGPRINT(RT_DEBUG_LOUD,("RTMPSendPacket --> pSrcBufVA == %p, SrcBufLen=%d\n", pSrcBufVA, SrcBufLen));

	// STEP 1. Decide number of fragments required to deliver this MSDU. 
	//	   The estimation here is not very accurate because difficult to 
	//	   take encryption overhead into consideration here. The result 
	//	   "NumberOfFrag" is then just used to pre-check if enough free 
	//	   TXD are available to hold this MSDU.


	if (*pSrcBufVA & 0x01)	// fragmentation not allowed on multicast & broadcast
		NumberOfFrag = 1;	 
	else if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED))
		NumberOfFrag = 1;	// Aggregation overwhelms fragmentation
	else if (CLIENT_STATUS_TEST_FLAG(&pAd->MacTab.Content[BSSID_WCID], fCLIENT_STATUS_AMSDU_INUSED))
		NumberOfFrag = 1;	// Aggregation overwhelms fragmentation
	else if ((pAd->CommonCfg.HTPhyMode.field.MODE == MODE_HTMIX) || (pAd->CommonCfg.HTPhyMode.field.MODE == MODE_HTGREENFIELD))
		NumberOfFrag = 1;	// MIMO RATE overwhelms fragmentation
	else
	{
		// The calculated "NumberOfFrag" is a rough estimation because of various 
		// encryption/encapsulation overhead not taken into consideration. This number is just
		// used to make sure enough free TXD are available before fragmentation takes place.
		// In case the actual required number of fragments of an NDIS packet 
		// excceeds "NumberOfFrag"caculated here and not enough free TXD available, the
		// last fragment (i.e. last MPDU) will be dropped in RTMPHardTransmit() due to out of 
		// resource, and the NDIS packet will be indicated NDIS_STATUS_FAILURE. This should 
		// rarely happen and the penalty is just like a TX RETRY fail. Affordable.
		
		AllowFragSize = (pAd->CommonCfg.FragmentThreshold) - LENGTH_802_11 - LENGTH_CRC;
		NumberOfFrag = ((PacketInfo.TotalPacketLength - LENGTH_802_3 + LENGTH_802_1_H) / AllowFragSize) + 1;
		// To get accurate number of fragmentation
		if (((PacketInfo.TotalPacketLength - LENGTH_802_3 + LENGTH_802_1_H) % AllowFragSize) == 0)
		{
			NumberOfFrag--;
		}
	}
	
	// Save fragment number to Ndis packet reserved field
	RTMP_SET_PACKET_FRAGMENTS(pPacket, NumberOfFrag);  
	// STEP 2. Check the requirement of RTS:
	//	   If multiple fragment required, RTS is required only for the first fragment
	//	   if the fragment size large than RTS threshold
	
	// Save RTS requirement to Ndis packet reserved field
	RTMP_SET_PACKET_RTS(pPacket, 0);
	RTMP_SET_PACKET_TXRATE(pPacket, pAd->CommonCfg.TxRate);

	//
	// STEP 3. Traffic classification. outcome = <UserPriority, QueIdx>
	//
	UserPriority = 0;
	QueIdx		 = QID_AC_BE;
	if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_WMM_INUSED))
	{
		USHORT Protocol;
		UCHAR  LlcSnapLen = 0, Byte0, Byte1;
		do
		{
			// get Ethernet protocol field
			Protocol = (USHORT)((pSrcBufVA[12] << 8) + pSrcBufVA[13]);
			if (Protocol <= 1500)
			{
				// get Ethernet protocol field from LLC/SNAP
				if (Sniff2BytesFromNdisBuffer(PacketInfo.pFirstBuffer, LENGTH_802_3 + 6, &Byte0, &Byte1) != NDIS_STATUS_SUCCESS)
					break;

				Protocol = (USHORT)((Byte0 << 8) + Byte1);
				LlcSnapLen = 8;
			}

			// always AC_BE for non-IP packet
			if (Protocol != 0x0800)
				break;

			// get IP header
			if (Sniff2BytesFromNdisBuffer(PacketInfo.pFirstBuffer, LENGTH_802_3 + LlcSnapLen, &Byte0, &Byte1) != NDIS_STATUS_SUCCESS)
				break;

			// return AC_BE if packet is not IPv4
			if ((Byte0 & 0xf0) != 0x40)
				break;

			UserPriority = (Byte1 & 0xe0) >> 5;
			QueIdx = MapUserPriorityToAccessCategory[UserPriority];

			// TODO: have to check ACM bit. apply TSPEC if ACM is ON
			// TODO: downgrade UP & QueIdx before passing ACM
			if (pAd->CommonCfg.APEdcaParm.bACM[QueIdx])
			{
				UserPriority = 0;
				QueIdx		 = QID_AC_BE;
			}
		} while (FALSE);
	}

	RTMP_SET_PACKET_UP(pPacket, UserPriority);

	NdisAcquireSpinLock(&pAd->TxSwQueueLock);
	if (pAd->TxSwQueue[QueIdx].Number >= MAX_PACKETS_IN_QUEUE)
	{
#ifdef BLOCK_NET_IF
		StopNetIfQueue(pAd, QueIdx, pPacket);
#endif // BLOCK_NET_IF //
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		NdisReleaseSpinLock(&pAd->TxSwQueueLock);
		return NDIS_STATUS_FAILURE;
	}
	else
	{
		InsertTailQueue(&pAd->TxSwQueue[QueIdx], PACKET_TO_QUEUE_ENTRY(pPacket));
	}
	NdisReleaseSpinLock(&pAd->TxSwQueueLock);

	pAd->RalinkCounters.OneSecOsTxCount[QueIdx]++; // TODO: for debug only. to be removed
	return NDIS_STATUS_SUCCESS;
}

/*
	========================================================================

	Routine Description:
		This subroutine will scan through releative ring descriptor to find
		out avaliable free ring descriptor and compare with request size.
		
	Arguments:
		pAd Pointer to our adapter
		QueIdx		Selected TX Ring
		
	Return Value:
		NDIS_STATUS_FAILURE 	Not enough free descriptor
		NDIS_STATUS_SUCCESS 	Enough free descriptor

	IRQL = PASSIVE_LEVEL
	IRQL = DISPATCH_LEVEL
	
	Note:
	
	========================================================================
*/
NDIS_STATUS RTMPFreeTXDRequest(
	IN		PRTMP_ADAPTER	pAd,
	IN		UCHAR			QueIdx,
	IN		UCHAR			NumberRequired,
	IN		PUCHAR			FreeNumberIs)
{
	ULONG		FreeNumber = 0;
	NDIS_STATUS 	Status = NDIS_STATUS_FAILURE;
			
	switch (QueIdx)
	{
		case QID_AC_BK:
		case QID_AC_BE:
		case QID_AC_VI:
		case QID_AC_VO:
		case QID_HCCA:
			if (pAd->TxRing[QueIdx].TxSwFreeIdx > pAd->TxRing[QueIdx].TxCpuIdx)
				FreeNumber = pAd->TxRing[QueIdx].TxSwFreeIdx - pAd->TxRing[QueIdx].TxCpuIdx - 1;
			else
				FreeNumber = pAd->TxRing[QueIdx].TxSwFreeIdx + TX_RING_SIZE - pAd->TxRing[QueIdx].TxCpuIdx - 1;

			if (FreeNumber >= NumberRequired)
				Status = NDIS_STATUS_SUCCESS;
			break;

		case QID_MGMT:
			if (pAd->MgmtRing.TxSwFreeIdx > pAd->MgmtRing.TxCpuIdx)
				FreeNumber = pAd->MgmtRing.TxSwFreeIdx - pAd->MgmtRing.TxCpuIdx - 1;
			else
				FreeNumber = pAd->MgmtRing.TxSwFreeIdx + MGMT_RING_SIZE - pAd->MgmtRing.TxCpuIdx - 1;

			if (FreeNumber >= NumberRequired)
				Status = NDIS_STATUS_SUCCESS;
			break;

		default:
			DBGPRINT(RT_DEBUG_ERROR,("RTMPFreeTXDRequest::Invalid QueIdx(=%d)\n", QueIdx));
			break;
	}
	*FreeNumberIs = (UCHAR)FreeNumber;

	return (Status);
}

VOID RTMPSendDisassociationFrame(
	IN	PRTMP_ADAPTER	pAd)
{
}
	
// IRQL = DISPATCH_LEVEL
VOID	RTMPSendNullFrame(
	IN	PRTMP_ADAPTER	pAd,
	IN	UCHAR			TxRate,
	IN	BOOLEAN 		bQosNull)
{
	UCHAR	NullFrame[48];
	ULONG	Length;
	PHEADER_802_11	pHeader_802_11;

    // WPA 802.1x secured port control
    if (((pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA) || 
         (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPAPSK) ||
         (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA2) || 
         (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA2PSK)
#ifdef WPA_SUPPLICANT_SUPPORT
			  || (pAd->StaCfg.IEEE8021X == TRUE)		
#endif 		
        ) &&
       (pAd->StaCfg.PortSecured == WPA_802_1X_PORT_NOT_SECURED)) 
	{
		return;
	}

	NdisZeroMemory(NullFrame, 48);
	Length = sizeof(HEADER_802_11);

	pHeader_802_11 = (PHEADER_802_11) NullFrame;
	
	pHeader_802_11->FC.Type = BTYPE_DATA;
	pHeader_802_11->FC.SubType = SUBTYPE_NULL_FUNC;
	pHeader_802_11->FC.ToDs = 1;
	COPY_MAC_ADDR(pHeader_802_11->Addr1, pAd->CommonCfg.Bssid);
	COPY_MAC_ADDR(pHeader_802_11->Addr2, pAd->CurrentAddress);
	COPY_MAC_ADDR(pHeader_802_11->Addr3, pAd->CommonCfg.Bssid);

	if (pAd->CommonCfg.bAPSDForcePowerSave)
	{
		pHeader_802_11->FC.PwrMgmt = PWR_SAVE;
	}
	else
	{
		pHeader_802_11->FC.PwrMgmt = (pAd->StaCfg.Psm == PWR_SAVE) ? 1: 0;
	}
	pHeader_802_11->Duration = pAd->CommonCfg.Dsifs + RTMPCalcDuration(pAd, TxRate, 14);

	pAd->Sequence++;
	pHeader_802_11->Sequence = pAd->Sequence;

	// Prepare QosNull function frame
	if (bQosNull)
	{
		pHeader_802_11->FC.SubType = SUBTYPE_QOS_NULL;
		
		// copy QOS control bytes
		NullFrame[Length]	=  0;
		NullFrame[Length+1] =  0;
		Length += 2;// if pad with 2 bytes for alignment, APSD will fail
	}
	
	MiniportMMRequest(pAd, 0, NullFrame, Length);
}

// IRQL = DISPATCH_LEVEL
VOID	RTMPSendRTSFrame(
	IN	PRTMP_ADAPTER	pAd,
	IN	PUCHAR			pDA,
	IN	unsigned int	NextMpduSize,
	IN	UCHAR			TxRate,
	IN	UCHAR			RTSRate,
	IN	USHORT			AckDuration,
	IN	UCHAR			QueIdx,
	IN	UCHAR			FrameGap)
{
}

/*
	========================================================================

	Routine Description:
		Copy frame from waiting queue into relative ring buffer and set 
	appropriate ASIC register to kick hardware encryption before really
	sent out to air.
		
	Arguments:
		pAd 	Pointer to our adapter
		PNDIS_PACKET	Pointer to outgoing Ndis frame
		NumberOfFrag	Number of fragment required
		
	Return Value:
		None

	IRQL = DISPATCH_LEVEL
	
	Note:
	
	========================================================================
*/
NDIS_STATUS 	RTMPHardTransmit(
	IN PRTMP_ADAPTER	pAd,
	IN PNDIS_PACKET 	pPacket,
	IN	UCHAR			QueIdx,
	OUT PULONG		pFreeTXDLeft)
{
	PACKET_INFO 	PacketInfo, NextPacketInfo;
	PUCHAR			pSrcBufVA, pNextPacketBufVA = 0;
	PUCHAR			pDMAHeaderBufVA, pDMAHeaderBufPtr;	
	UINT			SrcBytesCopied;
	UINT			FreeMpduSize, MpduSize = 0; 
	UINT			SrcBufLen, NextPacketBufLen = 0;
	UCHAR			SrcBufIdx;
	ULONG			SrcBufPA;
	INT 				SrcRemainingBytes;
	UCHAR			FrameGap = IFS_HTTXOP;
	HEADER_802_11	Header_802_11;
	PHEADER_802_11	pHeader80211 = NULL;
	PUCHAR			pExtraLlcSnapEncap; // NULL: no extra LLC/SNAP is required
	UCHAR			CipherAlg;
	PTXD_STRUC		pTxD;
#ifdef BIG_ENDIAN
	TXD_STRUC		TxD;
    PTXD_STRUC      pDestTxD;
#endif    
	PTXWI_STRUC 	pFirstTxWI;
	BOOLEAN 		bEAPOLFrame;
	BOOLEAN 		bAckRequired;
//	UCHAR			RetryMode = SHORT_RETRY;
	USHORT			AckDuration = 0;
	USHORT			EncryptionOverhead = 0;
	UCHAR			KeyIdx;
	PCIPHER_KEY 	pKey = NULL;
	UCHAR			PID = 0;
	PRTMP_TX_RING	pTxRing = &pAd->TxRing[QueIdx];
	PRTMP_SCATTER_GATHER_LIST	pScatterGatherList;
	RTMP_SCATTER_GATHER_LIST	LocalScatterGatherList;
	PNDIS_PACKET	pNextPacket = NULL;
	BOOLEAN 		bClonePacket = FALSE;
	UCHAR			MpduRequired, RtsRequired, UserPriority, MimoPs = MMPS_ENABLE;
	UCHAR			TxRate;
	//USHORT			Protocol;
	// To indicate cipher used for this packet
	NDIS_802_11_ENCRYPTION_STATUS	Cipher;
	//BOOLEAN 		bRTS_CTSFrame = FALSE;
	BOOLEAN 		bHTC = FALSE; // TRUE if HT AMSDU used for this MPDU.
	BOOLEAN 		bTXBA = FALSE; // TRUE if use block ack.
	ULONG		FreeTXDLeft = *pFreeTXDLeft;
	BOOLEAN 	TXWIAMPDU = FALSE;	   //AMPDU bit in TXWI
	UCHAR		TXWIBAWinSize  = 0; //BAWinSize field in TXWI
	ULONG		TxIdx =pAd->TxRing[QueIdx].TxCpuIdx;
	UCHAR		SDPNowUsed = 1;
	//UCHAR		ScatterNowUsed = 0;
	BOOLEAN 	IsSD0 = TRUE;
	PULONG		ptemp;
	BOOLEAN		letusprint = FALSE;
	UCHAR		RAWcid = BSSID_WCID;	// default in infra mode or adhoc mode using legacy rate, RA is BSSID. 
	UCHAR		RABAOriIdx = 0;	//The RA's BA Originator table index. 
	BOOLEAN		bPiggyBack = FALSE;
    BOOLEAN     bARALINK = FALSE;
	BOOLEAN		bAMSDU = FALSE;	// use HT AMSDU
	BOOLEAN		bAllowFrag = TRUE; // A-MPDU, A-MSDU, A-Ralink is not allowed to fragment
	BOOLEAN		bWMM = FALSE;
	BOOLEAN 	bDHCPFrame = FALSE;
	UCHAR 		AMSDUsubheader[14];
	UCHAR		padding = 0;
	BOOLEAN			bHTRate = FALSE;	// use HT AMSDU
	ULONG	OrgSrcBufPA;
	BOOLEAN		bWPA4WAYFrame = FALSE;

	if ((pAd->CommonCfg.bIEEE80211H == 1) && (pAd->CommonCfg.RadarDetect.RDMode != RD_NORMAL_MODE))
	{
		DBGPRINT(RT_DEBUG_INFO,("RTMPHardTransmit --> radar detect not in normal mode !!!\n"));
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return (NDIS_STATUS_FAILURE);
	}

	MpduRequired = RTMP_GET_PACKET_FRAGMENTS(pPacket);
	RtsRequired  = RTMP_GET_PACKET_RTS(pPacket);
	UserPriority = RTMP_GET_PACKET_UP(pPacket);
	TxRate		 = RTMP_GET_PACKET_TXRATE(pPacket);
	bWPA4WAYFrame 	= RTMP_GET_PACKET_WPA_4WAY_MSG(pPacket);
	
	RABAOriIdx = UserPriority;	// default in using UP as originator index in Infra mode. 
	
	// ---------------------------------------------
	// STEP 0. PARSING THE NDIS PACKET
	// ---------------------------------------------
	//
	// Prepare packet information structure which will be query for buffer descriptor
	//
	RTMP_QueryPacketInfo(pPacket, &PacketInfo, &pSrcBufVA, &SrcBufLen);

	if (SrcBufLen < 14)
	{
		DBGPRINT(RT_DEBUG_ERROR,("RTMPHardTransmit --> Ndis Packet buffer error !!!\n"));
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return (NDIS_STATUS_FAILURE);
	}

	if (pAd->bForcePrintTX == TRUE)
		letusprint = TRUE;
			
	// In HT rate adhoc mode, A-MPDU is often used. So need to lookup BA Table and MAC Entry. 
	// Note multicast packets in adhoc also use BSSID_WCID index. 
	if (ADHOC_ON(pAd) &&((*pSrcBufVA & 0x01) == 1))
	{
		RAWcid = MCAST_WCID;
	}
	else if (ADHOC_ON(pAd))		// only unicast data use Block Ack
	{
		MAC_TABLE_ENTRY *pEntry;
		pEntry = MacTableLookup(pAd, pSrcBufVA);
		if (pEntry)
		{
			RAWcid = (UCHAR)pEntry->Aid;
		}
	}

	if (pAd->MacTab.Content[RAWcid].HTPhyMode.field.MODE & 0x2)
	{
		bHTRate = TRUE;
		MimoPs = pAd->MacTab.Content[RAWcid].MmpsMode;
	}
	
	// Update Originator index as the RA's BA Originator index.
	RABAOriIdx = pAd->MacTab.Content[RAWcid].BAOriWcidArray[UserPriority];
	
	//
	// Only Broadcast packet could be DHCP or ARP.
	// If DHCP datagram or ARP datagram , we need to send it as Low rates.
	//
	if ((*pSrcBufVA & 0x01) && RTMPCheckDHCPFrame(pAd, pPacket))
	{
		bDHCPFrame = TRUE;
	}
	
	if (TxRate < pAd->CommonCfg.MinTxRate)
		TxRate = pAd->CommonCfg.MinTxRate;
		
#ifdef UNDER_CE
	bClonePacket = TRUE;		 
#else
#ifdef ME_98
	// 
	// It's not support ScatterGatherDma in Win9x and WinME
	// So, ClonePacket functionality is always TRUE.
	//
	bClonePacket = TRUE;	
#else
	if (RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_SCATTER_GATHER))
		bClonePacket = FALSE;  // cloning is not necessary if used scatter-gather DMA
	else 
		bClonePacket = TRUE;   // cloning is necessary when scatter-gather DMA failed.
#endif	// #if ME_98
#endif	// #ifdef UNDER_CE
	pNextPacket  = NULL;   // no aggregation is required

#if 1
	/* clone is not necessary in Linux or uCOS-II */
	bClonePacket = OS_Need_Clone_Packet();
#endif


	// check if aggregation applicable on this MSDU and the next one. If applicable
	// make sure both MSDUs use internally created NDIS PACKET structure so that both has
	// only one scatter-gather buffer
	// NOTE: 1.) aggregation not applicable when CKIP inused. because it's difficult for driver
	//		 to calculate CMIC on the aggregated MSDU
	//		 2.) used Aggregation if packet didn't be fragment, that means MpduRequired must be 1.
	//	3.) backward compatible SW Aggregation with RT2661, which is not MIMO txrate.
	//       4.) DHCP frame should not aggregate to increase success rate
	if (
		(          
		  //(OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_AGGREGATION_INUSED) && (pAd->CommonCfg.HTPhyMode.field.MODE == MODE_OFDM)) ||
		  //(OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_TX_AMSDU_INUSED))
		  (RAWcid != 0) && PeerIsAggreOn(pAd, TxRate, &pAd->MacTab.Content[RAWcid]) 
		 )  &&
		(pAd->TxSwQueue[QueIdx].Head != NULL)			   &&
		(FreeTXDLeft > 1)								 && //scattered-gather use many descriptor for a-msdu
		(pAd->StaCfg.bCkipOn == FALSE)						&&
		(MpduRequired == 1) 							   &&
        (bDHCPFrame == FALSE)									&&
		TxFrameIsAggregatible(pAd, NULL, pSrcBufVA))
	{
		PQUEUE_ENTRY pNextEntry;
//		UINT		TotalPacketLength;	// RT2860B . erase after verified.
		
		pNextEntry = RemoveHeadQueue(&pAd->TxSwQueue[QueIdx]);
		pNextPacket = QUEUE_ENTRY_TO_PACKET(pNextEntry);
		RTMP_QueryPacketInfo(pNextPacket, &NextPacketInfo, &pNextPacketBufVA, &NextPacketBufLen);
		
		DBGPRINT(RT_DEBUG_INFO,("AGGRE: 1st MSDU aggregatible(len=%d)\n", PacketInfo.TotalPacketLength));

		if (NextPacketBufLen < 14)
		{
			DBGPRINT(RT_DEBUG_ERROR,("RTMPHardTransmit -->	Ndis NextPacket buffer error !!!\n"));
			RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
			RELEASE_NDIS_PACKET(pAd, pNextPacket, NDIS_STATUS_FAILURE);
			return (NDIS_STATUS_FAILURE);
		}
		
		if (TxFrameIsAggregatible(pAd, pSrcBufVA, pNextPacketBufVA))
		{
		}
		else
		{
			// can't aggregate. put next packe back to TxSwQueue
			InsertHeadQueue(&pAd->TxSwQueue[QueIdx], pNextEntry);
			pNextPacket = NULL;  
		}
	}

	// Decide if Packet Cloning is required -
	// 1. when fragmentation && TKIP is inused, we have to clone it into a single buffer
	//	  NDIS_PACKET so that driver can pre-cluculate and append TKIP MIC at tail of the
	//	  source packet before hardware fragmentation can be performed
	// 2. when CKIP MIC format is inused, we have to walk through the source NDIS packet to
	//	  pre-calculate CMIC and insert at front of 1st output MPDU
	if ((RTMP_GET_PACKET_SOURCE(pPacket) == PKTSRC_NDIS) && (bClonePacket == FALSE))
	{
		pScatterGatherList = (PRTMP_SCATTER_GATHER_LIST) 
			GET_SG_LIST_FROM_PACKET(pPacket, &LocalScatterGatherList);
		ASSERT(pScatterGatherList);			
	}

	// Clone then release the original NDIS packet
	if (bClonePacket)
	{
		PNDIS_PACKET pOutPacket;
		NDIS_STATUS  Status;

		printk("### Buggy: clone a fragmented packet as a linear packet ###\n");

		Status = RTMPCloneNdisPacket(pAd, bAMSDU, pPacket, &pOutPacket);

		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_SUCCESS);

		if (Status != NDIS_STATUS_SUCCESS)
		{
			if (pNextPacket)
				RELEASE_NDIS_PACKET(pAd, pNextPacket, NDIS_STATUS_SUCCESS);
			return Status;
		}

		// Use the new cloned packet from now on
		pPacket = pOutPacket;
		RTMP_SET_PACKET_UP(pPacket, UserPriority);
		RTMP_QueryPacketInfo(pPacket, &PacketInfo, &pSrcBufVA, &SrcBufLen);
	}

    // if orginal Ethernet frame contains no LLC/SNAP, then an extra LLC/SNAP encap is required 
	EXTRA_LLCSNAP_ENCAP(pSrcBufVA, pExtraLlcSnapEncap);

	// link next MSDU into scatter list, if any
	if (pNextPacket)
	{
		//
		// A-MSDU or A-Ralink
		// 
		if (CLIENT_STATUS_TEST_FLAG(&pAd->MacTab.Content[BSSID_WCID], fCLIENT_STATUS_AMSDU_INUSED))
		{
            USHORT subMSDULen;

			if (pExtraLlcSnapEncap)
				subMSDULen = SrcBufLen - LENGTH_802_3 + LENGTH_802_1_H;
			else
				subMSDULen = SrcBufLen - LENGTH_802_3;

			// 2.1. Make sub-AMSDU header,			
			RTMPMoveMemory(&AMSDUsubheader[0], pSrcBufVA, 12);

			AMSDUsubheader[12] = (subMSDULen & 0xFF00) >> 8;
			AMSDUsubheader[13]= subMSDULen & 0xFF;
			// This is a A-MSDU frame
			bAMSDU = TRUE;
		}
		else
		{
			bARALINK = TRUE;// Ralink Aggregation 
			pAd->RalinkCounters.OneSecTxAggregationCount++;
		}

		bAllowFrag = FALSE;

		//
		// set local scatter gather list 
		// 
		LocalScatterGatherList.NumberOfElements = 2;
		LocalScatterGatherList.Elements[1].Length = NextPacketBufLen;
		LocalScatterGatherList.Elements[1].Address = pNextPacketBufVA;
		PacketInfo.TotalPacketLength += NextPacketBufLen;
	}

	pScatterGatherList = &LocalScatterGatherList;

	// ------------------------------------------
	// STEP 0.1 Add 802.1x protocol check.
	// ------------------------------------------
	
	if (NdisEqualMemory(EAPOL, pSrcBufVA + 12, 2))
	{
		bEAPOLFrame = TRUE;
	}
	else
		bEAPOLFrame = FALSE;

	// we using lowest basic rate when DHCP and EAPOL packet
	if ((bDHCPFrame == TRUE) || (bEAPOLFrame == TRUE))
		bHTRate = FALSE;

	//
	// WPA 802.1x secured port control - drop all non-802.1x frame before port secured
	//
	if (((pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA) || 
		 (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPAPSK) ||
		 (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA2) || 
		 (pAd->StaCfg.AuthMode == Ndis802_11AuthModeWPA2PSK)
#ifdef WPA_SUPPLICANT_SUPPORT
		  || (pAd->StaCfg.IEEE8021X == TRUE) 
#endif // WPA_SUPPLICANT_SUPPORT //     		 
		  ) &&
		((pAd->StaCfg.PortSecured == WPA_802_1X_PORT_NOT_SECURED) || (pAd->StaCfg.MicErrCnt >= 2)) &&
		(bEAPOLFrame == FALSE))
	{    
		DBGPRINT(RT_DEBUG_TRACE,("RTMPHardTransmit --> Drop packet before port secured !!!\n"));
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		if (pNextPacket)
		{
			RELEASE_NDIS_PACKET(pAd, pNextPacket, NDIS_STATUS_FAILURE);
		}

		return (NDIS_STATUS_FAILURE);
	}

	// allow to use A-MPDU and A-MSDU altogether.
	if (((pAd->MacTab.Content[RAWcid].TXBAbitmap&(1<<UserPriority)) != 0) 
		&& (((INFRA_ON(pAd))) || ((*pSrcBufVA & 0x01) == 0))
        && (bDHCPFrame == FALSE))		// only unicast RA use Block Ack.
	{
		bTXBA = TRUE;
		bAllowFrag = FALSE;

	}	

	if (pAd->CommonCfg.AckPolicy[QueIdx] != NORMAL_ACK)
	{
		bAckRequired = FALSE;
		PID = 0;
		Cipher = pAd->StaCfg.PairCipher; // Cipher for Unicast
	}
	else if (bTXBA)
	{
		bAckRequired = TRUE;
		Cipher = pAd->StaCfg.PairCipher; // Cipher for Unicast
	}
	else if ((*pSrcBufVA & 0x01) && (ADHOC_ON(pAd))) // Multicast or Broadcast
	{
		bAckRequired = FALSE;
		INC_COUNTER64(pAd->WlanCounters.MulticastTransmittedFrameCount);
		Cipher = pAd->StaCfg.GroupCipher; // Cipher for Multicast or Broadcast
	}
	else
	{
		bAckRequired = TRUE;
		PID = 0;
		Cipher = pAd->StaCfg.PairCipher; // Cipher for Unicast
		if (letusprint == TRUE)
			DBGPRINT_RAW(RT_DEBUG_TRACE, ("Use PairCipher = %d ", Cipher));
	}

	// When HT rate, always use IFS_HTTXOP
	if (bHTRate == TRUE)
		FrameGap = IFS_HTTXOP;
	// When RTS enable, always use IFS_HTTXOP, or ASIC will not help send RTS frame
	else if (PacketInfo.TotalPacketLength > pAd->CommonCfg.RtsThreshold)
		FrameGap = IFS_HTTXOP;
	// When B/G protection enable, always use IFS_HTTXOP, or ASIC will not help send CTS frame
	else if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_BG_PROTECTION_INUSED))
		FrameGap = IFS_HTTXOP;
	else if(OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_WMM_INUSED))
	{
		if (!pAd->CommonCfg.APEdcaParm.Txop[QueIdx])
			FrameGap = IFS_HTTXOP;
		else
			FrameGap = IFS_BACKOFF;
	}
	// Tx burst mode
	else if (pAd->CommonCfg.bEnableTxBurst)
		FrameGap = IFS_HTTXOP;
	// otherwise, always BACKOFF before transmission
	else
		FrameGap = IFS_BACKOFF; 	// Default frame gap mode

	// ------------------------------------------------------------------
	// STEP 1. WAKE UP PHY
	//		outgoing frame always wakeup PHY to prevent frame lost and 
	//		turn off PSM bit to improve performance
	// ------------------------------------------------------------------
	// not to change PSM bit, just send this frame out?
	if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_DOZE))
		AsicForceWakeup(pAd);
	// It should not change PSM bit, when APSD turn on.
	if ((!(pAd->CommonCfg.bAPSDCapable && pAd->CommonCfg.APEdcaParm.bAPSDCapable) && (pAd->CommonCfg.bAPSDForcePowerSave == FALSE)) || bEAPOLFrame)
	{
		if (pAd->StaCfg.Psm == PWR_SAVE)
			MlmeSetPsmBit(pAd, PWR_ACTIVE);
	}

	// -----------------------------------------------------------------
	// STEP 2. MAKE A COMMON 802.11 HEADER SHARED BY ENTIRE FRAGMENT BURST. Fill sequence later. 
	// -----------------------------------------------------------------

	MAKE_802_11_HEADER(pAd, Header_802_11, pSrcBufVA);
	if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_WMM_INUSED))
	{
		Header_802_11.FC.SubType = SUBTYPE_QDATA;
		bWMM = TRUE;

		if ((pAd->MacTab.Content[BSSID_WCID].HTCapability.ExtHtCapInfo.PlusHTC == TRUE) && (pAd->MacTab.Content[BSSID_WCID].HTPhyMode.field.MODE >= MODE_HTMIX) && (bDHCPFrame == FALSE))
		{
			Header_802_11.FC.Order = 1;	// HTC bit share with order bit
			bHTC = TRUE;
		}
	}

	// --------------------------------------------------------
	// STEP 3. FIND ENCRYPT KEY AND DECIDE CIPHER ALGORITHM
	//		Find the WPA key, either Group or Pairwise Key
	//		LEAP + TKIP also use WPA key.
	// --------------------------------------------------------
	// Decide WEP bit and cipher suite to be used. Same cipher suite should be used for whole fragment burst
	// In Cisco CCX 2.0 Leap Authentication
	//		   WepStatus is Ndis802_11Encryption1Enabled but the key will use PairwiseKey
	//		   Instead of the SharedKey, SharedKey Length may be Zero.
	KeyIdx = 0xff;
	if (bEAPOLFrame)
	{
		ASSERT(pAd->SharedKey[BSS0][0].CipherAlg <= CIPHER_CKIP128);
		
		// 4-way handshaking frame must be clear
		if (!(bWPA4WAYFrame) && (pAd->SharedKey[BSS0][0].CipherAlg) &&
			(pAd->SharedKey[BSS0][0].KeyLen))
		{
			CipherAlg = pAd->SharedKey[BSS0][0].CipherAlg;
			KeyIdx = 0;
		}
	}
	else if (Cipher == Ndis802_11Encryption1Enabled)
	{
#ifdef LEAP_SUPPORT		
		if (pAd->StaCfg.CkipFlag & 0x10) // Cisco CKIP KP is on
		{
			if (LEAP_CCKM_ON(pAd))
			{
				if (Header_802_11.Addr1[0] & 0x01)
					KeyIdx = 1;
				else
					KeyIdx = 0;
			}
			else
				KeyIdx = pAd->StaCfg.DefaultKeyId;
		}
		else if (pAd->StaCfg.CkipFlag & 0x08) // only CKIP CMIC
			KeyIdx = pAd->StaCfg.DefaultKeyId;
		else if (LEAP_CCKM_ON(pAd))
		{
			if (Header_802_11.Addr1[0] & 0x01)
				KeyIdx = 1;
			else
				KeyIdx = 0;
		}
		else	// standard WEP64 or WEP128
#endif // LEAP_SUPPORT //		
			KeyIdx = pAd->StaCfg.DefaultKeyId;
	}
	else if ((Cipher == Ndis802_11Encryption2Enabled) ||
			 (Cipher == Ndis802_11Encryption3Enabled))
	{
		if (Header_802_11.Addr1[0] & 0x01) // multicast
			KeyIdx = pAd->StaCfg.DefaultKeyId;
		else if (pAd->SharedKey[BSS0][0].KeyLen)
			KeyIdx = 0;
		else
			KeyIdx = pAd->StaCfg.DefaultKeyId;
	}

	if (KeyIdx == 0xff)
		CipherAlg = CIPHER_NONE;
	else if ((Cipher == Ndis802_11EncryptionDisabled) || (pAd->SharedKey[BSS0][KeyIdx].KeyLen == 0))
		CipherAlg = CIPHER_NONE;
#ifdef WPA_SUPPLICANT_SUPPORT
    else if ( pAd->StaCfg.WpaSupplicantUP && 
             (Cipher == Ndis802_11Encryption1Enabled) &&
             (pAd->StaCfg.IEEE8021X == TRUE) &&
             (pAd->StaCfg.PortSecured == WPA_802_1X_PORT_NOT_SECURED))
        CipherAlg = CIPHER_NONE;
#endif // WPA_SUPPLICANT_SUPPORT //
	else
	{
		Header_802_11.FC.Wep = 1;
		CipherAlg = pAd->SharedKey[BSS0][KeyIdx].CipherAlg;
		pKey = &pAd->SharedKey[BSS0][KeyIdx];
	}


	// STEP 3.1 if TKIP is used and fragmentation is required. Driver has to
	//			append TKIP MIC at tail of the scatter buffer (This must be the
	//			ONLY scatter buffer in the NDIS PACKET). 
	//			MAC ASIC will only perform IV/EIV/ICV insertion but no TKIP MIC
	if ((MpduRequired > 1) && (CipherAlg == CIPHER_TKIP))
	{
		ASSERT(pScatterGatherList->NumberOfElements == 1);

		RTMPCalculateMICValue(pAd, pPacket, pExtraLlcSnapEncap, pKey, 0);

		NdisMoveMemory(pSrcBufVA + SrcBufLen, pAd->PrivateInfo.Tx.MIC, 8);
		SrcBufLen += 8;
		PacketInfo.TotalPacketLength += 8;
		CipherAlg = CIPHER_TKIP_NO_MIC;
	}
	else
	{
		//
		// Snice the ScatterGather may allocate on the contiguous range of the base physical address
		// In that case, the pScatterGatherList->Elements[0].Length will equal to PacketInfo.TotalPacketLength
		// And the pScatterGatherList->NumberOfElements will be reduce one snice the contiguous range of the data.
		// So we need to reset the SrcBufLen as pScatterGatherList->Elements[0].Length as our first Packet length.
		// Otherwise will not get the next pScatterGatherList Elements's length and cause TxRing Full.
		// This is ok on PKTSRC_NDIS case or not the PKTSRC_NDIS case.
		//	
		ASSERT(pScatterGatherList);
		SrcBufLen		  = pScatterGatherList->Elements[0].Length;
	}
	
	// ----------------------------------------------------------------
	// STEP 4. Make RTS frame or CTS-to-self frame if required
	// ----------------------------------------------------------------

	//
	// calcuate the overhead bytes that encryption algorithm may add. This
	// affects the calculate of "duration" field
	//
	if ((CipherAlg == CIPHER_WEP64) || (CipherAlg == CIPHER_WEP128)) 
		EncryptionOverhead = 8; //WEP: IV[4] + ICV[4];
	else if (CipherAlg == CIPHER_TKIP_NO_MIC)
		EncryptionOverhead = 12;//TKIP: IV[4] + EIV[4] + ICV[4], MIC will be added to TotalPacketLength
	else if (CipherAlg == CIPHER_TKIP)
		EncryptionOverhead = 20;//TKIP: IV[4] + EIV[4] + ICV[4] + MIC[8]
	else if (CipherAlg == CIPHER_AES)
		EncryptionOverhead = 16;	// AES: IV[4] + EIV[4] + MIC[8]
	else
		EncryptionOverhead = 0;

	// decide how much time an ACK/CTS frame will consume in the air
	AckDuration = RTMPCalcDuration(pAd, pAd->CommonCfg.ExpectedACKRate[TxRate], 14);

	if (RtsRequired)
	{
		// Leave this because maybe need driver send RTS/CTS future.
	}

	// --------------------------------------------------------
	// STEP 5. SEARCH BA PEER(s). As Sta, always use MacTab.Content[BSSID_WCID]
	// --------------------------------------------------------
{
    if (RAWcid != MCAST_WCID) 
	{
		MAC_TABLE_ENTRY *pMacEntry;

		pMacEntry = &pAd->MacTab.Content[RAWcid];
		
        if (bTXBA)
        {	
            RABAOriIdx = pMacEntry->BAOriWcidArray[UserPriority];
            TXWIBAWinSize = pAd->BATable.BAOriEntry[RABAOriIdx].BAWinSize;
            
            TXWIAMPDU = TRUE;
            Header_802_11.Frag = 0;
        }
        Header_802_11.Sequence = pMacEntry->TxSeq[UserPriority];
        pMacEntry->TxSeq[UserPriority] = (pMacEntry->TxSeq[UserPriority]+1) & 0x0FFF;
    }
    else
    {
		Header_802_11.Sequence = pAd->Sequence;
		pAd->Sequence = (pAd->Sequence+1) & 0x0FFFF; // next sequence 		
    }
}
	
	// --------------------------------------------------------
	// STEP 6. START MAKING MPDU(s)
	//		Start Copy Ndis Packet into Ring buffer.
	//		For frame required more than one ring buffer (fragment), all ring buffers
	//		have to be filled before kicking start tx bit.
	//		Make sure TX ring resource won't be used by other threads
	// --------------------------------------------------------
	
	SrcBufIdx		  = 0;
	SrcRemainingBytes = PacketInfo.TotalPacketLength;

	{
		ULONG	OffsetSrcVA = (ULONG)pScatterGatherList->Elements[0].Address;

		OffsetSrcVA 	+= LENGTH_802_3;	// skip 802.3 header
		SrcBufLen		-= LENGTH_802_3;	// skip 802.3 header										
		SrcBufPA		= (ULONG)PCI_MAP_SINGLE(pAd, (char *)OffsetSrcVA,
							SrcBufLen, PCI_DMA_TODEVICE);

		OrgSrcBufPA = SrcBufPA - LENGTH_802_3;
		// update accurate payload size
		SrcRemainingBytes = PacketInfo.TotalPacketLength - LENGTH_802_3;
	}

	//NdisAcquireSpinLock(&pAd->TxRingLock);

	//TxIdx =pAd->TxRing[QueIdx].TxCpuIdx;

	// Update SW_TX_IDX[QueIdx] and *pFreeTXDLeft after finish the copy while-loop
	// FreeTXDLeft is used to record current Free TXD.	If return middle way, real *pFreeTXDLeft don't change.
	FreeTXDLeft = *pFreeTXDLeft;
	
	// Outer while loop loops for fragment packet. At beginning of outer loop, it's a complete wireless MPDU.
	do
	{
		ULONG  BufBasePaLow;

		if (FreeTXDLeft == 0)
		{
			DBGPRINT(RT_DEBUG_ERROR,("RTMPHardTransmit: No Free TXD Left\n"));
			pAd->RalinkCounters.TxRingErrCount++;
			//NdisReleaseSpinLock(&pAd->TxRingLock);
			RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
			if (pNextPacket)
			{
				RELEASE_NDIS_PACKET(pAd, pNextPacket, NDIS_STATUS_FAILURE);
			}
			return (NDIS_STATUS_RESOURCES);
		}
		
		//
		// STEP 6.1 And 1ST DMA BUFFER
		//
		//Save FirstTxWI.  Need to fill MPDUtotlabytecount when finish 
		pFirstTxWI = (PTXWI_STRUC)pTxRing->Cell[TxIdx].DmaBuf.AllocVa;

		pDMAHeaderBufVA = (PUCHAR) pTxRing->Cell[TxIdx].DmaBuf.AllocVa;		 
		BufBasePaLow  = RTMP_GetPhysicalAddressLow (pTxRing->Cell[TxIdx].DmaBuf.AllocPa);
		// need to fill MPDU total byte count field at last step
		pDMAHeaderBufPtr = pDMAHeaderBufVA;

		//
		// STEP 6.2 COPY TXWI and 802.11 HEADER INTO 1ST DMA BUFFER. IV/EIV will set by ASIC after searching on-chip WC table
		//
		NdisZeroMemory(pFirstTxWI, TXWI_SIZE);
		
		// bWIV==FALSE, then Driver doesn't need to fill IV/EIV, which is filled by FCE.
		//as Sta, the RA always is BSSID. So always use BSSID_WCID.
		// DHCP uses low rate
		if ((bDHCPFrame == TRUE) || (bEAPOLFrame == TRUE))
		{
			RTMPWriteTxWI(pAd, pFirstTxWI,  FALSE, FALSE, FALSE, TXWIAMPDU, bAckRequired, FALSE, 
				TXWIBAWinSize, RAWcid, 0,PID, UserPriority,  TxRate, 
				IFS_BACKOFF, bPiggyBack, &pAd->MacTab.Content[MCAST_WCID].HTPhyMode);
			pAd->LastTxRate = (USHORT)(pAd->MacTab.Content[MCAST_WCID].HTPhyMode.word);

			DBGPRINT_RAW(RT_DEBUG_INFO,("RTMPHardTransmit-DHCP packet uses MCS=%d, STBC=%d, ShortGI=%d, Mode=%d, BW=%d \n",
				pAd->MacTab.Content[MCAST_WCID].HTPhyMode.field.MCS, pAd->MacTab.Content[MCAST_WCID].HTPhyMode.field.STBC, pAd->MacTab.Content[MCAST_WCID].HTPhyMode.field.ShortGI,
				pAd->MacTab.Content[MCAST_WCID].HTPhyMode.field.MODE, pAd->MacTab.Content[MCAST_WCID].HTPhyMode.field.BW));
		}
		else if ((ADHOC_ON(pAd)) && (RAWcid > BSSID_WCID))	// for different adhoc STA, we use the same HTPhyMode as (RAWcid=BSSID_WCID)
		{
			RTMPWriteTxWI(pAd, pFirstTxWI,  FALSE, FALSE, FALSE, TXWIAMPDU, bAckRequired, FALSE, 
				TXWIBAWinSize, RAWcid, 0,PID, UserPriority,  TxRate, 
				FrameGap, bPiggyBack, &pAd->MacTab.Content[BSSID_WCID].HTPhyMode);
			pAd->LastTxRate = (USHORT)(pAd->MacTab.Content[BSSID_WCID].HTPhyMode.word);
		}
		else
		{
			RTMPWriteTxWI(pAd, pFirstTxWI,  FALSE, FALSE, FALSE, TXWIAMPDU, bAckRequired, FALSE, 
				TXWIBAWinSize, RAWcid, 0,PID, UserPriority,  TxRate, 
				FrameGap, bPiggyBack, &pAd->MacTab.Content[RAWcid].HTPhyMode);
			pAd->LastTxRate = (USHORT)(pAd->MacTab.Content[RAWcid].HTPhyMode.word);
		}

		pDMAHeaderBufPtr += TXWI_SIZE;
		NdisMoveMemory(pDMAHeaderBufPtr, &Header_802_11, sizeof(Header_802_11));

		pDMAHeaderBufPtr		+= sizeof(Header_802_11);
		MpduSize = sizeof(Header_802_11);
					
		//
		// STEP 6.3 ACQUIRE TXD. Init Ndispacket = NULL. Because it TxD is not necessarily the lastsec
		//
#ifndef BIG_ENDIAN
		pTxD  = (PTXD_STRUC) pTxRing->Cell[TxIdx].AllocVa;
#else
		pDestTxD  = (PTXD_STRUC)pTxRing->Cell[TxIdx].AllocVa;
		TxD = *pDestTxD;
		pTxD = &TxD;
#endif
        
		RING_PACKET_INIT(pTxRing, TxIdx);
		NdisZeroMemory(pTxD, TXD_SIZE);
		pTxD->DMADONE = 1;
		ptemp = (PULONG)pTxD;
//		RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);

		// 1st Buf always has TXWI+802.11header. This is preallocated DMABuf. But we need to reassign it every beginning of MPDU.
		//pTxD->SDPtr0 = BufBasePaLow;

		FreeTXDLeft--;

		//
		// Fragmentation is not allowed on multicast & broadcast. and is nether used in HT rate.
		// So, we need to used the MAX_FRAG_THRESHOLD instead of pAd->CommonCfg.FragmentThreshold
		// otherwise if PacketInfo.TotalPacketLength > pAd->CommonCfg.FragmentThreshold then
		// packet will be fragment on multicast & broadcast.
		//
		if ((*pSrcBufVA & 0x01) || (bAllowFrag == FALSE)) //(pAd->CommonCfg.HTPhyMode.field.MODE >= MODE_HTMIX))
		{
			FreeMpduSize = MAX_AGGREGATION_SIZE - sizeof(Header_802_11) - LENGTH_CRC;
		}
		else
		{
			FreeMpduSize = pAd->CommonCfg.FragmentThreshold - sizeof(Header_802_11) - LENGTH_CRC;
		}

		// Note: if HT rate, WMM must inused.
		//if (OPSTATUS_TEST_FLAG(pAd, fOP_STATUS_WMM_INUSED))
		if (bWMM)
		{
			// HT rate must has QOS CONTROL bytes
			if (bTXBA)
				*pDMAHeaderBufPtr		  =  (UserPriority & 0x0f) ;
			else
				*pDMAHeaderBufPtr		  =  (UserPriority & 0x0f) | pAd->CommonCfg.AckPolicy[QueIdx];

			if (bAMSDU == TRUE)
			{
				*pDMAHeaderBufPtr |= 0x80;
			}

			*(pDMAHeaderBufPtr+1)	  =  0;
			pDMAHeaderBufPtr		  += 2;
			MpduSize += 2;
			FreeMpduSize  -= 2;

			// Pad 2 bytes to make 802.11 header 4-byte alignment
			if (bHTC == TRUE)
			{
				// HTC Control field(4 bytes) is following QOS field, padding 2 bytes
				NdisZeroMemory(pDMAHeaderBufPtr, 6);
				if (pAd->CommonCfg.bRdg == TRUE)
					*(pDMAHeaderBufPtr+4)|=0x80;
				if (pAd->bLinkAdapt == TRUE)
				{	//bit2MRQ=1. Request for MCS feedback
					*(pDMAHeaderBufPtr) |=0x4;
					// Set identifier as 2
					*(pDMAHeaderBufPtr) |= ((pAd->MRS)<<3);
				}
				pDMAHeaderBufPtr		  += 6;
				MpduSize += 4;

			}
			else
			{
			    // padding 2 bytes
				NdisZeroMemory(pDMAHeaderBufPtr, 2);
				pDMAHeaderBufPtr		  += 2;
			}
		}

		if (pNextPacket && (bAMSDU == FALSE))
		{
			FreeMpduSize  = MAX_AGGREGATION_SIZE;
            //software aggregation puts the 2nd MSDU length(extra 2-byte field) after QOS_CONTROL. little endian
		    // Handle Aggregation (A-MSUD for 11n station in HTMode or A-Ralink in Legacy 11g Ralink station, )
			if (bARALINK)
            {         
			    // For RA Aggregation, 
			    // put the 2nd MSDU length(extra 2-byte field) after QOS_CONTROL in little endian format
				*pDMAHeaderBufPtr		  =  (UCHAR)NextPacketBufLen & 0xff;
				*(pDMAHeaderBufPtr+1)	  =  (UCHAR)(NextPacketBufLen >> 8);
                
				pDMAHeaderBufPtr		  += 2;
                MpduSize += 2;
				Header_802_11.FC.Order = 1; // steal "order" bit to mark "aggregation"
			}
		}

        if (bAMSDU)
		{
			/* A-MSDU format: DA + SA + Length + MSDU + Padding */
     		//
     		// use 1st dma buffer to store subframe header 
     		// DA(6)+SA(6)+Len(2)
     		// 							
     		// build subframe header
     		NdisMoveMemory(pDMAHeaderBufPtr, &AMSDUsubheader[0], 14);
     		pDMAHeaderBufPtr += 14;
			MpduSize += 14;
		}
        
		//
		// STEP 5.4 COPY LLC/SNAP, CKIP MIC INTO 1ST DMA BUFFER ONLY WHEN THIS 
		//			MPDU IS THE 1ST OR ONLY FRAGMENT 
		//
		if (Header_802_11.Frag == 0)
		{
			//if (pExtraLlcSnapEncap && (bAMSDU == FALSE))
			if (pExtraLlcSnapEncap)
			{
#ifdef CONFIG_CKIP_SUPPORT
				if ((pAd->StaCfg.CkipFlag & 0x08) && (bEAPOLFrame == FALSE))
				{
					// CKIP_LLC_SNAP is inused. The frame format must be -
					//	   802.11 header + CKIP_LLCSNAP(8) + MIC(4) + TxSEQ(4) + Proto(2) + Payload
					NdisMoveMemory(pDMAHeaderBufPtr, CKIP_LLC_SNAP, sizeof(CKIP_LLC_SNAP));
					pDMAHeaderBufPtr += sizeof(CKIP_LLC_SNAP);
					//
					// 1.) Insert MIC [4]
					//
					RTMPCkipInsertCMIC(pAd, pDMAHeaderBufPtr, (PUCHAR)&Header_802_11, pPacket, pKey, CKIP_LLC_SNAP);
					pDMAHeaderBufPtr += 4 ;
					//
					// 2.) Insert TX Sequence.
					//
					NdisMoveMemory(pDMAHeaderBufPtr, pAd->StaCfg.TxSEQ, 4);
					pDMAHeaderBufPtr += 4;
					NdisMoveMemory(pDMAHeaderBufPtr, pSrcBufVA + 12, 2);
					pDMAHeaderBufPtr += 2;
					MpduSize += 10;
					FreeMpduSize -= (sizeof(CKIP_LLC_SNAP) + 10);

					// Update TxSEQ for next TX
					{ 
						int i = 3;
						pAd->StaCfg.TxSEQ[i] = pAd->StaCfg.TxSEQ[i] + 2;
						while (pAd->StaCfg.TxSEQ[i] == 0x00)
						{
							i--;
							if (i < 0) break;
							pAd->StaCfg.TxSEQ[i] ++;
						}
					}

				}
				else
#endif /* CONFIG_CKIP_SUPPORT */
				{
					// Insert LLC-SNAP encapsulation
					NdisMoveMemory(pDMAHeaderBufPtr, pExtraLlcSnapEncap, 6);
					pDMAHeaderBufPtr += 6;
					NdisMoveMemory(pDMAHeaderBufPtr, pSrcBufVA + 12, 2);
					pDMAHeaderBufPtr += 2;
					MpduSize += LENGTH_802_1_H;
					FreeMpduSize -= LENGTH_802_1_H;
				}
			}
		}

		//
		// set 1st header size into TX DESC
		//   --- TxWI + 802.11 Header
		//   --- if (!A-MSDU)
		// 			+ LLC/SNAP Encap
		// 		 else
		// 			+ DA(6)+SA(6)+Len(2)
		// 		
		pTxD->SDPtr0 = BufBasePaLow;
		pTxD->SDLen0 = (ULONG)(pDMAHeaderBufPtr - pDMAHeaderBufVA);

		//
		// STEP 5.5 TRAVERSE NDIS PACKET TO BUILD THE MPDU PAYLOAD
		//
		SrcBytesCopied	 = 0;

		// Inner while loop is for getting enough TXD to send this MPDU. 
		// SDPNowUsed starts from 1 because WI and 802.11Header already use SDP0 of 1st TXD.
		SDPNowUsed = 1;
		IsSD0 = TRUE;
		do
		{
			// Assign Currently used TxD.  
			SDPNowUsed ++;
			if (SrcBufLen <= FreeMpduSize)
			{
				// scatter-gather buffer still fit into current MPDU
				if (SrcBufLen != 0)
				{
					switch (SDPNowUsed%2) 
					{
						case 1:  
{
							PUCHAR pSubHeader;
                            UCHAR   SDLen0 = 0;

							ASSERT(FreeTXDLeft);
							ASSERT(bAMSDU);
							ASSERT(pNextPacket);

							FreeTXDLeft--;
							INC_RING_INDEX(TxIdx, TX_RING_SIZE);
							//
							// AMSDU packet: 2nd subframe
							// 
							// Init TxRing
#ifndef BIG_ENDIAN
							pTxD = (PTXD_STRUC) pTxRing->Cell[TxIdx].AllocVa;
#else
							//Swap the endian of old pTxD and prepare to handle next pTxD.
							RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
							WriteBackToDescriptor((PUCHAR)pDestTxD, (PUCHAR)pTxD, FALSE, TYPE_TXD);
		
							pDestTxD  = (PTXD_STRUC)pTxRing->Cell[TxIdx].AllocVa;
							TxD = *pDestTxD;
							pTxD = &TxD;
#endif
                            
							NdisZeroMemory(pTxD, TXD_SIZE);
							pTxD->DMADONE = 1;
							//RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);

							//
							// use 1st dma buffer to store sub-frame header
							// pervious sub-frame padding + DA(6) + SA(6) + Len(2)
							// 
							BufBasePaLow = RTMP_GetPhysicalAddressLow(pTxRing->Cell[TxIdx].DmaBuf.AllocPa);
							pSubHeader = pTxRing->Cell[TxIdx].DmaBuf.AllocVa;

                            if (bAMSDU)
                            {
    							ASSERT(padding < 4);
    							pSubHeader += padding;

    							// build sub-frame header
    							NdisMoveMemory(pSubHeader, pNextPacketBufVA, 12);
                                // if orginal Ethernet frame contains no LLC/SNAP, then an extra LLC/SNAP encap is required 
                                EXTRA_LLCSNAP_ENCAP(pNextPacketBufVA, pExtraLlcSnapEncap);
                                SDLen0 = padding + 14;
                                NextPacketBufLen -= LENGTH_802_3;

    							if (pExtraLlcSnapEncap)
    							{
    								NextPacketBufLen += LENGTH_802_1_H;
    								NdisMoveMemory(pSubHeader+14, pExtraLlcSnapEncap, 6);
    							    NdisMoveMemory(pSubHeader+14+6, pNextPacketBufVA+12, 2); 								
    								SDLen0 += LENGTH_802_1_H;
    							}
                                
    							pSubHeader[12] = (NextPacketBufLen & 0xFF00) >> 8;
    							pSubHeader[13] = NextPacketBufLen & 0xFF;
                            }
                            else if (bARALINK)
                            {
                                NdisMoveMemory(pSubHeader , pNextPacketBufVA, 12);
								NdisMoveMemory(pSubHeader+12, pNextPacketBufVA+12, 2);

								SDLen0 = LENGTH_802_3;

								//Here we didn't need to subtract the VLAN tag length, because we do it in previous.
								NextPacketBufLen -= LENGTH_802_3;
                            }
							// update MPDU size
							MpduSize += SDLen0;

							// build TX Desc
							pTxD->SDPtr0 = BufBasePaLow;
							pTxD->SDLen0 = SDLen0;
							pTxD->SDPtr1 = SrcBufPA;
							pTxD->SDLen1 = SrcBufLen;

                            //RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);
							//pTxD->DMADONE = 0;
							break;
}
						case 0: 
							pTxD->SDPtr1 = SrcBufPA;
							pTxD->SDLen1 = SrcBufLen;							

							if (bAMSDU || bARALINK)
							{							
                                pTxRing->Cell[TxIdx].pNdisPacket = pPacket;
    							pTxRing->Cell[TxIdx].pNextNdisPacket = NULL;
                                RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);
							} 
							else
							{
								ASSERT(OrgSrcBufPA != SrcBufPA);
							}

							IsSD0 = FALSE;
							if (letusprint == TRUE)
							{
								ptemp = (PULONG)pTxD;
								DBGPRINT(RT_DEBUG_TRACE,("HardTx TxD[%ld]  SDPtr1 =%d,  SDLen1 = %08lx: %08lx: %08lx: %08lx  \n", TxIdx, pTxD->SDLen1,  *ptemp, *(ptemp+1), *(ptemp+2), *(ptemp+3)));
							}
							break;
						default:	// should never happen
							break;

					}
				}
				else
				{
					SDPNowUsed--;	// not use this segment.
				} 
				 

				SrcBytesCopied	+= SrcBufLen;
				//pDMAHeaderBufPtr			+= SrcBufLen;
				FreeMpduSize	-= SrcBufLen;
				MpduSize		+= SrcBufLen;
				SrcBufPA		+= SrcBufLen;
				//SrcBufLen		 = 0;


				//
				// calcuate padding bytes for sub-frame
				// put the padding bytes into 1st DMA buffer of next Tx DESC
				// 
				if (bAMSDU && (SrcBufIdx == 0))
				{
					ULONG AlignLen;

					AlignLen = ((14+SrcBufLen)+3) & ~(0x3);
					padding = AlignLen - (14+SrcBufLen);
				}

				// advance to next scatter-gather BUFFER
				SrcBufIdx++;

                if (SrcBufIdx < pScatterGatherList->NumberOfElements)
				{
					pTxD->LastSec0 = 0;
					pTxD->LastSec1 = 0;

					{
						ULONG	OffsetSrcVA = (ULONG) pScatterGatherList->Elements[SrcBufIdx].Address;
						//UCHAR   *Header = (UCHAR *)OffsetSrcVA;

						SrcBufLen = pScatterGatherList->Elements[SrcBufIdx].Length;					

						OffsetSrcVA      += LENGTH_802_3;  // skip 802.3 header
						SrcBufLen        -= LENGTH_802_3;  // skip 802.3 header
					    
						SrcBufPA          = PCI_MAP_SINGLE(pAd, (char *)OffsetSrcVA,
												SrcBufLen, PCI_DMA_TODEVICE);  
						SrcRemainingBytes -= LENGTH_802_3;
					} 
				}
				else 
				{
					// Set Last Data Segment
					pTxD->LastSec0 = 0;
					pTxD->LastSec1 = 1;
					SrcBufLen = 0;
				}
				if (SrcBufLen == 0)
					break;
			}
			else
			{
				//scatter-gather buffer exceed current MPDU. leave some of the buffer to next MPDU
				switch (SDPNowUsed%2) 
				{
					case 1: 		// Init TxRing 
							NdisZeroMemory(pTxD, sizeof(pTxD));
							pTxD->SDPtr0 = SrcBufPA;
							pTxD->SDLen0 = FreeMpduSize;
							pTxD->LastSec0 = 1;
							DBGPRINT(RT_DEBUG_INFO,("Last Fraged TXD. Use pTxD[] SDPtr0, SDLen0 = %d \n",pTxD->SDLen0));

							break;
					case 0: 
							pTxD->SDPtr1 = SrcBufPA;
							pTxD->SDLen1	 = FreeMpduSize;
							pTxD->LastSec1 = 1;
							pTxD->DMADONE = 0;

							DBGPRINT(RT_DEBUG_INFO,("Last Fraged TXD. Use pTxD[] SDPtr1,  SDLen1 = %d \n", pTxD->SDLen1));

							break;
					default:	// should never happen
							break;
				}
				SrcBytesCopied += FreeMpduSize;
				pDMAHeaderBufPtr		   += FreeMpduSize;
				MpduSize	   += FreeMpduSize;
				SrcBufPA	   += FreeMpduSize;
				SrcBufLen	   -= FreeMpduSize;

				// a complete MPDU is built. break out to write TXD of this MPDU
				break;
			}
		}	while (TRUE);		// End of copying payload
		
		// remaining size of the NDIS packet payload
		SrcRemainingBytes -= SrcBytesCopied;

		//
		// STEP 5.6 MODIFY MORE_FRAGMENT BIT & DURATION FIELD. WRITE TXD
		//
		pHeader80211 = (PHEADER_802_11)((PUCHAR)pFirstTxWI + TXWI_SIZE);
		if (SrcRemainingBytes > 0) // more fragment is required
		{
			ULONG NextMpduSize;

			pHeader80211->FC.MoreFrag = 1;
			NextMpduSize = min((ULONG)SrcRemainingBytes, (ULONG)pAd->CommonCfg.FragmentThreshold);

			if (NextMpduSize < pAd->CommonCfg.FragmentThreshold)
			{
				// In this case, we need to include LENGTH_802_11 and LENGTH_CRC for calculating Duration.
				pHeader80211->Duration = (3 * pAd->CommonCfg.Dsifs) + 
								(2 * AckDuration) + 
								RTMPCalcDuration(pAd, TxRate, NextMpduSize + EncryptionOverhead + LENGTH_802_11 + LENGTH_CRC);
			}
			else
			{
				pHeader80211->Duration = (3 * pAd->CommonCfg.Dsifs) + 
								(2 * AckDuration) + 
								RTMPCalcDuration(pAd, TxRate, NextMpduSize + EncryptionOverhead);
			}

			//
			// On fragment case, set burst mode to TRUE to tell ASIC to continue fetch the fragment data.
			// that belongs to the same MSDU.
			//
			RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);
			pTxD->Burst = 1;
			// Although WIV is filled by FCE, byte count need driver to fill for ASIC initialization
			pFirstTxWI->MPDUtotalByteCount = MpduSize;
			pFirstTxWI->txop = IFS_SIFS;
			Header_802_11.Frag ++;	 // increase Frag #
			
			pTxRing->Cell[TxIdx].pNdisPacket = NULL;
            pTxRing->Cell[TxIdx].pNextNdisPacket = NULL;
		}
		else // this is the last or only fragment
		{
			pHeader80211->FC.MoreFrag = 0;
			if (pHeader80211->Addr1[0] & 0x01) // multicast/broadcast
				pHeader80211->Duration = 0;
			else
				pHeader80211->Duration = pAd->CommonCfg.Dsifs + AckDuration;

			if ((bEAPOLFrame) && (TxRate > RATE_6) && (TxRate <= RATE_54))
				TxRate = RATE_6;

			// record packet to free later
			if (bAMSDU || bARALINK)
                // TxIdx already increased above (in case 1)
                // And pPacket = pTxRing->Cell[TxIdx - 1].pNdisPacket above (in case 0)
                pTxRing->Cell[TxIdx].pNdisPacket = pNextPacket; 
            else
			pTxRing->Cell[TxIdx].pNdisPacket = pPacket; 			 
            
   			pTxRing->Cell[TxIdx].pNextNdisPacket = NULL;
			pFirstTxWI->MPDUtotalByteCount = MpduSize;
            RTMPWriteTxDescriptor(pAd, pTxD, FALSE, FIFO_EDCA);
		}

		// Inform MAC ASIC TKIP engine this is a fragment, so that TKIP MIC is appended by driver at the last fragment
		pFirstTxWI->FRAG = ((pHeader80211->FC.MoreFrag) || (Header_802_11.Frag != 0));
		ASSERT(MpduSize < 4096);

		if (letusprint == TRUE)
		{
			DBGPRINT(RT_DEBUG_INFO,("(TxRate=%d, TxRate=%d, Frag=%d) \n", TxRate, pAd->CommonCfg.TxRate, MpduRequired));
			ptemp = (PULONG)pTxD;
			DBGPRINT(RT_DEBUG_TRACE,("HardTx TxD[%ld]= %08lx: %08lx: %08lx: %08lx  \n", TxIdx, *ptemp, *(ptemp+1), *(ptemp+2), *(ptemp+3)));
			ptemp = (PULONG)pFirstTxWI;
			DBGPRINT(RT_DEBUG_TRACE,("HardTx TxWI= %08lx: %08lx: %08lx: %08lx  \n", *ptemp, *(ptemp+1), *(ptemp+2), *(ptemp+3)));
			ptemp = (PULONG)pHeader80211;
			DBGPRINT(RT_DEBUG_TRACE,("HardTx Header= %08lx: %08lx: %08lx: %08lx: %08lx: %08lx: %08lx: %08lx: %08lx  \n", *ptemp, *(ptemp+1), *(ptemp+2), *(ptemp+3), *(ptemp+4), *(ptemp+5), *(ptemp+6), *(ptemp+7), *(ptemp+8)));
		}
		
#ifdef BIG_ENDIAN
		// First check if we need to do endian conversion for the TX_WI.
		RTMPWIEndianChange((PUCHAR)pFirstTxWI, TYPE_TXWI);

		// Do endian conversion for the 802.11 header
		RTMPFrameEndianChange(pAd, (PUCHAR)(pDMAHeaderBufVA+TXWI_SIZE), DIR_WRITE, FALSE);
		
		// Do endian converison for the Tx Descriptor.
		RTMPDescriptorEndianChange((PUCHAR)pTxD, TYPE_TXD);
		WriteBackToDescriptor((PUCHAR)pDestTxD, (PUCHAR)pTxD, FALSE, TYPE_TXD);
#endif
        
		pAd->RalinkCounters.KickTxCount++;
		INC_RING_INDEX(TxIdx, TX_RING_SIZE);
	} while (SrcRemainingBytes > 0);

	// Success , Update True FreeTXDLeft
	*pFreeTXDLeft = FreeTXDLeft;
	// Update  SW_TX_IDX
	pAd->TxRing[QueIdx].TxCpuIdx = TxIdx;
	// Make sure to release Tx ring resource

	RTMP_IO_WRITE32(pAd, TX_CTX_IDX0 + QueIdx * 0x10 , pAd->TxRing[QueIdx].TxCpuIdx);

	//NdisReleaseSpinLock(&pAd->TxRingLock);

	// the NDIS PACKET "pPacket" will be released at DMA done interrupt service routine
	return (NDIS_STATUS_SUCCESS);
}

// this portion already move to cmm_data.c
// RTMPSuspendMsduTransmission,
// RTMPResumeMsduTransmission

/*
	========================================================================

	Routine Description:
		Check Rx descriptor, return NDIS_STATUS_FAILURE if any error dound
		
	Arguments:
		pRxD		Pointer to the Rx descriptor
		
	Return Value:
		NDIS_STATUS_SUCCESS 	No err
		NDIS_STATUS_FAILURE 	Error
		
	Note:
	
	========================================================================
*/
NDIS_STATUS RTMPCheckRxError(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PHEADER_802_11	pHeader,
	IN	PRXWI_STRUC pRxWI,
	IN	PRXD_STRUC		pRxD)
{
	PCIPHER_KEY pWpaKey;
	INT dBm;
	
	// Phy errors & CRC errors
	if (/*(pRxD->PhyErr) ||*/ (pRxD->Crc))
	{
		// Check RSSI for Noise Hist statistic collection.
		dBm = (INT) (pRxWI->RSSI0) - pAd->BbpRssiToDbmDelta;
		if (dBm <= -87)
			pAd->StaCfg.RPIDensity[0] += 1;
		else if (dBm <= -82)
			pAd->StaCfg.RPIDensity[1] += 1;
		else if (dBm <= -77)
			pAd->StaCfg.RPIDensity[2] += 1;
		else if (dBm <= -72)
			pAd->StaCfg.RPIDensity[3] += 1;
		else if (dBm <= -67)
			pAd->StaCfg.RPIDensity[4] += 1;
		else if (dBm <= -62)
			pAd->StaCfg.RPIDensity[5] += 1;
		else if (dBm <= -57)
			pAd->StaCfg.RPIDensity[6] += 1;
		else if (dBm > -57)
			pAd->StaCfg.RPIDensity[7] += 1;
		
		return(NDIS_STATUS_FAILURE);
	}
		
	// Add Rx size to channel load counter, we should ignore error counts
	pAd->StaCfg.CLBusyBytes += (pRxD->SDL0 + 14);

	// Drop ToDs promiscous frame, it is opened due to CCX 2 channel load statistics
	if (pHeader != NULL)
		if (pHeader->FC.ToDs)
			return(NDIS_STATUS_FAILURE);

	// Drop not U2M frames, cant's drop here because we will drop beacon in this case
	// I am kind of doubting the U2M bit operation
	// if (pRxD->U2M == 0)
	//	return(NDIS_STATUS_FAILURE);

	// drop decyption fail frame
	if (pRxD->CipherErr)
	{
		if (pRxD->CipherErr == 2)
			{DBGPRINT_RAW(RT_DEBUG_TRACE,("pRxD ERROR: ICV ok but MICErr "));}
		else if (pRxD->CipherErr == 1)
			{DBGPRINT_RAW(RT_DEBUG_TRACE,("pRxD ERROR: ICV Err "));}
		else if (pRxD->CipherErr == 3)
			DBGPRINT_RAW(RT_DEBUG_TRACE,("pRxD ERROR: Key not valid "));
		
		DBGPRINT_RAW(RT_DEBUG_TRACE,(" %d (len=%d, Mcast=%d, MyBss=%d, Wcid=%d, KeyId=%d)\n", 
			pRxD->CipherErr,
			pRxD->SDL0, 
			pRxD->Mcast | pRxD->Bcast,
			pRxD->MyBss, 
			pRxWI->WirelessCliID, 
//			CipherName[pRxD->CipherAlg],
			pRxWI->KeyIndex));

		//
		// MIC Error
		//
		if (pRxD->CipherErr == 2)
		{			
			pWpaKey = &pAd->SharedKey[BSS0][pRxWI->KeyIndex];
#ifdef WPA_SUPPLICANT_SUPPORT                                    
            if (pAd->StaCfg.WpaSupplicantUP)
                WpaSendMicFailureToWpaSupplicant(pAd, 
                                   (pWpaKey->Type == PAIRWISEKEY) ? TRUE:FALSE);
            else
#endif // WPA_SUPPLICANT_SUPPORT //                             
			    RTMPReportMicError(pAd, pWpaKey);

			DBGPRINT_RAW(RT_DEBUG_ERROR,("Rx MIC Value error\n"));
		}
		
		if (pHeader == NULL)
			return(NDIS_STATUS_SUCCESS);
		/*if ((pRxD->CipherAlg == CIPHER_AES) &&
			(pHeader->Sequence == pAd->FragFrame.Sequence))
		{
			//
			// Acceptable since the First FragFrame no CipherErr problem.
			//
			return(NDIS_STATUS_SUCCESS);
		}*/
		
		return(NDIS_STATUS_FAILURE);
	}
	
	return(NDIS_STATUS_SUCCESS);
}

/*
	========================================================================

	Routine Description:
		Apply packet filter policy, return NDIS_STATUS_FAILURE if this frame
		should be dropped.
		
	Arguments:
		pAd 	Pointer to our adapter
		pRxD			Pointer to the Rx descriptor
		pHeader 		Pointer to the 802.11 frame header
		
	Return Value:
		NDIS_STATUS_SUCCESS 	Accept frame
		NDIS_STATUS_FAILURE 	Drop Frame
		
	IRQL = DISPATCH_LEVEL
	
	Note:
		Maganement frame should bypass this filtering rule.
	
	========================================================================
*/
NDIS_STATUS RTMPApplyPacketFilter(
	IN	PRTMP_ADAPTER	pAd, 
	IN	PRXD_STRUC		pRxD, 
	IN	PHEADER_802_11	pHeader)
{
	
	// 0. Management frame should bypass all these filtering rules.
	if (pHeader == NULL)
		return(NDIS_STATUS_SUCCESS);

	if (pHeader->FC.Type == BTYPE_MGMT)
		return(NDIS_STATUS_SUCCESS);
	
	// 0.1	Drop all Rx frames if MIC countermeasures kicks in
	if (pAd->StaCfg.MicErrCnt >= 2)
	{
		//DBGPRINT_RAW(RT_DEBUG_TRACE,("Rx dropped by MIC countermeasure\n"));
		return(NDIS_STATUS_FAILURE);
	}
	
	// 1. Drop unicast to me packet if NDIS_PACKET_TYPE_DIRECTED is FALSE
	if (pRxD->U2M) 
	{
		if (!RX_FILTER_TEST_FLAG(pAd, fRX_FILTER_ACCEPT_DIRECT))
		{
			DBGPRINT_RAW(RT_DEBUG_TRACE,("Rx U2M dropped by RX_FILTER\n"));
			return(NDIS_STATUS_FAILURE);
		}
	}
		
	// 2. Drop broadcast packet if NDIS_PACKET_TYPE_BROADCAST is FALSE
	else if (pRxD->Bcast)
	{
		if (!RX_FILTER_TEST_FLAG(pAd, fRX_FILTER_ACCEPT_BROADCAST))
		{
			DBGPRINT_RAW(RT_DEBUG_TRACE,("Rx BCAST dropped by RX_FILTER\n"));
			return(NDIS_STATUS_FAILURE);
		}
	}
			
	// 3. Drop (non-Broadcast) multicast packet if NDIS_PACKET_TYPE_ALL_MULTICAST is false
	//	  and NDIS_PACKET_TYPE_MULTICAST is false.
	//	  If NDIS_PACKET_TYPE_MULTICAST is true, but NDIS_PACKET_TYPE_ALL_MULTICAST is false.
	//	  We have to deal with multicast table lookup & drop not matched packets.
	else if (pRxD->Mcast)
	{
		if (!RX_FILTER_TEST_FLAG(pAd, fRX_FILTER_ACCEPT_ALL_MULTICAST))
		{
			if (!RX_FILTER_TEST_FLAG(pAd, fRX_FILTER_ACCEPT_MULTICAST))
			{
				DBGPRINT_RAW(RT_DEBUG_TRACE,("Rx MCAST dropped by RX_FILTER\n"));
				return(NDIS_STATUS_FAILURE);
			}
			else
			{
				if (pHeader == NULL)
				{
					DBGPRINT_RAW(RT_DEBUG_TRACE,("Rx  MCAST pHeader == NULL\n"));
					return(NDIS_STATUS_SUCCESS);
				}
				// Selected accept multicast packet based on multicast table
				/*for (i = 0; i < pAd->NumberOfMcastAddresses; i++)
				{
					if (MAC_ADDR_EQUAL(pHeader->Addr1, pAd->McastTable[i]))
						break;		// Matched
				}

				// Not matched
				if (i == pAd->NumberOfMcastAddresses)
				{
					DBGPRINT(RT_DEBUG_TRACE,("Rx MCAST %02x:%02x:%02x:%02x:%02x:%02x dropped by RX_FILTER\n",
						pHeader->Addr1[0], pHeader->Addr1[1], pHeader->Addr1[2], 
						pHeader->Addr1[3], pHeader->Addr1[4], pHeader->Addr1[5]));
					return(NDIS_STATUS_FAILURE);
				}
				else
				{
					DBGPRINT(RT_DEBUG_TRACE,("Accept multicast %02x:%02x:%02x:%02x:%02x:%02x\n",
						pHeader->Addr1[0], pHeader->Addr1[1], pHeader->Addr1[2], 
						pHeader->Addr1[3], pHeader->Addr1[4], pHeader->Addr1[5]));
				}*/
			}
		}
	}

	// 4. Not U2M, not Mcast, not Bcast, must be unicast to other DA.
	//	  Since we did not implement promiscuous mode, just drop this kind of packet for now.
	else
	{
		DBGPRINT_RAW(RT_DEBUG_TRACE,("Not U2M, not Mcast, not BcastL\n"));
		return(NDIS_STATUS_FAILURE);
	}
	return(NDIS_STATUS_SUCCESS);	
}

// this portion already move to cmm_data.c
// RTMPCheckTxSwQueue,
// RTMPReportMicError,

		
// This portion already move to rt_os.c
// RTMPCloneNdisPacket,
// RTMPAllocateNdisPacket,
// RTMPFreeNdisPacket

	
// this portion already move to cmm_data.c
// RTMPFreeTXDUponTxDmaDone,
// RTMPCheckDHCPFrame

ULONG  HashBytesPolynomial(UCHAR *value, unsigned int len)
{
   unsigned char *word = value;
   unsigned int ret = 0; 
   unsigned int i;	
   
   for(i=0; i < len; i++)
   {
	  int mod = i % 32;
	  ret ^=(unsigned int) (word[i]) << mod;
	  ret ^=(unsigned int) (word[i]) >> (32 - mod);
   }
   return ret;
}

